from core_10x.rc import RC

if __name__ == '__main__':
    #import os;assert os.environ.setdefault('UI_PLATFORM', 'Rio') == os.getenv('UI_PLATFORM')

    from ui_10x.utils import ux, ux_pick_date


    app = ux.init()

    ux_pick_date(on_accept=lambda value: print(value) or RC(True))

    print(ux_pick_date())




from core_10x.basket import Basket, Bucketizer
from core_10x.code_samples.person import Person
from core_10x.xinf import XInf
from core_10x.named_constant import NamedCallable

class FEATURE(NamedCallable):
    A60     = lambda p: p.older_than(60)
    W200    = lambda p: p.weight >= 200

if __name__ == '__main__':

    all_people = Person.load_many()

    class AGGREGATOR(NamedCallable):
        WEIGHT     = lambda gen: sum(v*q for v, q in gen)

    b = Basket(base_class = Person, aggregator_class = AGGREGATOR)

    #bz = Bucketizer.by_feature(Person, FEATURE.W200)
    #bz = Bucketizer.by_breakpoints(Person, lambda p: p.weight, -XInf, 100, 170, 200, XInf)
    #bz = Bucketizer.by_range(Person, Person.T.weight, ('underweight', -XInf, 100), ('normal', 170., 180.), ['overweight', 190., XInf])
    #bz = Bucketizer.by_feature(Person, lambda p: p.weight, 190., 200.)

    bz = Bucketizer.by_range(Person, Person.T.weight, ('underweight', -XInf, 100), ('normal', 170., 180.), ['overweight', 190., XInf])
    bz2 = Bucketizer.by_feature(Person, Person.T.age)

    b.bucketizers = [bz, bz2]
    for i, p in enumerate(all_people):
        b.add(p, i)

    for tag, bucket in b.tags_buckets():
        print(f'tag = {tag}:')
        for member, qty in bucket.members_qtys():
            print(f'{member} = {qty}')

    r = b.weight
    print(f'weight = {r}')

    b.bucketizers = []
    for i, p in enumerate(all_people[:-1]):
        b.add(p, i)

    r = b.weight
    print(f'weight = {r}')

    bz = Bucketizer.by_range(Person, Person.T.age, ('normal', 50, XInf ))
    b.add_bucketizer(bz)

    r = b.weight
    print(f'weight = {r}')


    bz = Bucketizer.by_range(Person, Person.T.weight, ('underweight', -XInf, 100), ('normal', 170., 180.), ['overweight', 190., XInf])
    b.bucketizers = [bz]
    for i, p in enumerate(all_people):
        b.add(p, i)

    print('*** by subtags ***')
    # for tag, bucket in b.tags_buckets():
    #     print(f'tag = {tag}: {bucket.data}')
    gen = b.buckets_by_subtags('overweight')
    for tag, bucket in gen:
        print(f'tag = {tag}:')
        print(bucket.data)



from rio.debug.monkeypatches import apply_monkeypatches

import rio
from ui_10x.rio.components.splitter import Splitter


# Example app to demonstrate the Splitter component
class SplitterApp(rio.Component):
    def build(self) -> rio.Component:
        splitter = Splitter(
            children=[
                rio.Text('Pane 1', style='heading3'),
                rio.Text('Pane 2', style='heading3'),
                rio.Rectangle(content=rio.Container(rio.Container(rio.FlowContainer()))),
            ],
            handle_size=0.3,  # Splitter handle width
            direction='horizontal',
        )
        return rio.Container(rio.Column(rio.Column(rio.Row(rio.Button('A'), rio.Button('B')), rio.Container(splitter))))


# Run the app
if __name__ == '__main__':
    app = rio.App(build=SplitterApp)
    apply_monkeypatches()
    # app._run_in_window(debug_mode=True)
    app._run_as_web_server(debug_mode=True, port=8080, host='')



import inspect
from PyQt6.QtCore import QAbstractTableModel, QModelIndex, Qt, QVariant
from PyQt6.QtWidgets import QAbstractScrollArea, QTableView

from core_10x.traitable import Trait, Traitable

from ui_10x.table_header_view import HeaderModel, HeaderView
from ui_10x.traitable_editor import TraitableEditor
from ui_10x.traitable_view import TraitableView
from ui_10x.utils import UxStyleSheet, ux_text_alignment


class Model(QAbstractTableModel):
    def __init__(self, model: HeaderModel, data, parent = None):
        super().__init__(parent = parent)
        self.model = model
        self.m_data = data
        self.col2name = { i: name for i, name in enumerate( model.traits ) }
        self.name2col = { name: i for i, name in self.col2name.items() }

    def trait_name(self, col: int) -> str:
        return self.col2name.get(col)

    def rowCount(self, parent = None, *args, **kwargs):
        return len(self.m_data)

    def columnCount( self, parent = None, *args, **kwargs ):
        if not self.m_data:
            return 0
        return len(self.col2name)

    #-- TODO: comment out when HeaderView is working
    def headerData(self, section, orientation, role = None):
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                return list(self.model.header_structure.values())[section]

        return None

    def data(self, index: QModelIndex, role: int = None):
        if role == Qt.ItemDataRole.UserRole:
            return self.model

        entity: Traitable = self.m_data[index.row()]
        trait_name = self.col2name.get(index.column())

        cls = entity.__class__
        trait: Trait = cls.trait(trait_name)
        if role == Qt.ItemDataRole.DisplayRole and index.isValid():
            if trait:
                value = entity.get_value(trait)
                text = trait.to_str(value)
            else:
                text = ''
            return text

        if role == Qt.ItemDataRole.ForegroundRole:
            sh = entity.get_style_sheet(trait)
            return UxStyleSheet.fg_color(sh)

        if role == Qt.ItemDataRole.BackgroundRole:
            sh = entity.get_style_sheet(trait)
            return UxStyleSheet.bg_color(sh)

        if role == Qt.ItemDataRole.TextAlignmentRole:
            return ux_text_alignment(trait.ui_hint.param('align_h', 0), horizontal = True)

        return QVariant()

    def set_data(self, i: QModelIndex, dummy, role = None) -> bool:
        self.dataChanged.emit(i, i)
        return True

    def change_data(self, row: int, trait_name: str):
        col = self.name2col.get(trait_name)
        #i = QModelIndex(row, col)
        i = self.index(row, col)
        self.set_data(i, None, role = None)

    def insertRows(self, row: int, num_rows: int, parent = None, *args, **kwargs) -> bool:
        #self.beginInsertRows( QModelIndex(), row, row + num_rows )
        return True

    def extendData(self, data) -> bool:
        """
        :param data: an iterable of entities
        :return: True if model reset happened as well
        """
        if data:
            the_data = self.m_data
            is_empty = not bool(the_data)
            if is_empty:
                self.beginResetModel()

            first = len(the_data)
            last = first + len(data) - 1
            self.beginInsertRows(QModelIndex(), first, last)
            the_data.extend(data)
            self.insertRows(first, len(data))
            self.endInsertRows()

            if is_empty:
                self.endResetModel()
            return is_empty

    def entity(self, row: int) -> Traitable:
        return self.m_data[row]

class TableView(QTableView):
    def __init__(self, entities_or_class, view: TraitableView = None):
        assert entities_or_class, 'first arg must not be empty'

        if inspect.isclass(entities_or_class):
            assert issubclass(entities_or_class, Traitable), 'first arg is a class, but not a subclass of Traitable'
            proto = entities_or_class
            entities = []
        else:
            entities = entities_or_class
            proto = entities[0].__class__

        if not view:
            view = TraitableView.default(proto)

        super().__init__()
        self.wheel_enabled = True

        hv = HeaderView(view, parent = self)
        hv.setStretchLastSection(True)

        self.setHorizontalHeader(hv)
        model = Model(hv.m_model, entities, self)
        self.setModel(model)

        self.setAlternatingRowColors(True)
        self.resizeColumnsToContents()

        self.setSizeAdjustPolicy(QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)

    def wheelEvent(self, *args, **kwargs):
        if self.wheel_enabled:
            QTableView.wheelEvent(self, args[0])

    def keyPressEvent(self, *args, **kwargs):
        if self.wheel_enabled:
            QTableView.keyPressEvent(self, args[0])

    def mousePressEvent(self, mouse_event):
        pos = mouse_event.position()    #-- QPointF
        row = self.rowAt(int(pos.y()))
        col = self.columnAt(int(pos.x()))
        if row == -1 or col == -1:
            return

        entity = self.model().entity(row)
        trait_name = self.model().trait_name(col)
        if not trait_name:
            return

        editor_class: type[TraitableEditor] = TraitableEditor.find_editor_class(entity.__class__)     #-- no alternative packages, look under ui subdir only
        mouse_btn = mouse_event.button()
        if mouse_btn == Qt.MouseButton.LeftButton:
            cb = editor_class.left_mouse_callback(trait_name)
        elif mouse_btn == Qt.MouseButton.RightButton:
            cb = editor_class.right_mouse_callback(trait_name)
        else:
            return

        if cb:
            cb(self, entity)
        else:
            super().mousePressEvent(mouse_event)

    def render_traitable(self, row: int, entity: Traitable):
        last_col = len(self.model().col2name) - 1
        first_index = self.model().createIndex(row, 0)
        last_index = self.model().createIndex(row, last_col)
        self.model().dataChanged.emit(first_index, last_index)

    def extend_data(self, data):
        if self.model().extendData(data):
            self.resizeColumnsToContents()
            #self.resizeRowsToContents()



from __future__ import annotations

from core_10x.code_samples.person import Person

import rio
from examples import UserSessionContext
from examples import components as comps


def guard(event: rio.GuardEvent) -> str | None:
    return None if event.session[UserSessionContext].authenticated else '/'


@rio.page(
    name='CollectionEditor',
    url_segment='ce',
    guard=guard,
)
class CollectionEditor(rio.Component):
    def build(self) -> rio.Component:
        return comps.CollectionEditorComponent(collection_class=Person)



from core_10x.traitable_heir import TraitableHeir, T, RT

class Conventions(TraitableHeir):
    mkt_name: str   = T(T.ID)

    days_to_settle: int = T()

c_dir = dict(Conventions.s_dir)

class XConventions(Conventions):
    days_to_settle_twice: int   = RT()

    def days_to_settle_twice_get(self):
        return 2*self.days_to_settle

c_dir2 = Conventions.s_dir

if __name__ == '__main__':
    from core_10x.manual_tests.traitable_heir_test import Conventions, XConventions

    mkt_name = 'WTI NYMEX'
    c = Conventions(mkt_name = mkt_name)
    c.days_to_settle = 2
    print(c.days_to_settle)

    loc = 'ICE'
    mkt_name2 = f'{mkt_name} {loc}'
    c2 = Conventions(mkt_name = mkt_name2)
    c2._grantor = c

    print(c2.days_to_settle)

    c.days_to_settle = 3
    print(c2.days_to_settle)

    xc = XConventions(mkt_name = mkt_name)
    xc._grantor = c

    print(xc.days_to_settle_twice)


from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from py10x_kernel import BTraitableClass

    from core_10x.trait import Trait

# ===================================================================================================================================
#
#   age = LT(value)
#   name = 'Sasha'      #-- name = EQ('Sasha')
#   weight = BETWEEN(170, 180, bounds = (True, False)
#   weight =
# ===================================================================================================================================


# fmt: off
class _mongo_label:
    EQ      = '$eq'
    NE      = '$ne'
    GT      = '$gt'
    GE      = '$gte'
    LT      = '$lt'
    LE      = '$lte'
    IN      = '$in'
    NIN     = '$nin'
    REGEX   = '$regex'
    AND     = '$and'
    OR      = '$or'
    #NOT     = '$not'
# fmt: on

LABEL = _mongo_label


class _filter(ABC):
    @abstractmethod
    def eval(self, left_value) -> bool: ...
    @abstractmethod
    def prefix_notation(self, trait: Trait = None, traitable_class: BTraitableClass = None) -> dict: ...


class Op(_filter, ABC):
    label = ''

    def __init_subclass__(cls, label: str = None):
        if label is None:
            label = getattr(LABEL, cls.__name__)
        cls.label = label

    def __new__(cls, expression=None):
        obj = super().__new__(cls)
        obj.right_value = expression
        return obj

    def serialize_right_value(self, trait: Trait = None, traitable_class: BTraitableClass = None) -> dict:
        return trait.serialize_value(self.right_value, replace_xnone=True) if trait and traitable_class else self.right_value

    def prefix_notation(self, trait: Trait = None, traitable_class: BTraitableClass = None) -> dict:
        # noinspection PyTypeChecker
        return {self.label: self.serialize_right_value(trait, traitable_class)}


class NOT_EMPTY(Op, label=''):
    def prefix_notation(self, trait: Trait = None, traitable_class: BTraitableClass = None) -> dict:
        raise NotImplementedError

    def eval(self, left_value) -> bool:
        return bool(left_value)


class EQ(Op):
    def eval(self, left_value) -> bool:
        return left_value == self.right_value


class NE(Op):
    def eval(self, left_value) -> bool:
        return left_value != self.right_value


class GT(Op):
    def eval(self, left_value) -> bool:
        return left_value > self.right_value


class GE(Op):
    def eval(self, left_value) -> bool:
        return left_value >= self.right_value


class LT(Op):
    def eval(self, left_value) -> bool:
        return left_value < self.right_value


class LE(Op):
    def eval(self, left_value) -> bool:
        return left_value <= self.right_value


class IN(Op):
    def __new__(cls, values: list | tuple):
        assert isinstance(values, list) or isinstance(values, tuple), f'{cls.__name__}() requires a list or tuple'
        return super().__new__(cls, values)

    def serialize_right_value(self, trait: Trait = None, traitable_class: BTraitableClass = None) -> dict:
        return [trait.serialize_value(value, replace_xnone=True) for value in self.right_value] if trait and traitable_class else self.right_value

    def eval(self, left_value) -> bool:
        return left_value in self.right_value


class NIN(IN):
    def eval(self, left_value) -> bool:
        return left_value not in self.right_value


# class REGEX(Op):


class BETWEEN(Op, label=''):
    def __new__(cls, a, b, bounds=(True, True)):
        obj = super().__new__(cls)
        assert isinstance(bounds, tuple) and len(bounds) == 2, f'{cls.__name__} - (bool, bool) is expected for bounds'

        bound_a, bound_b = bounds
        obj.left = GE(a) if bound_a else GT(a)
        obj.right = LE(b) if bound_b else LT(b)
        return obj

    def eval(self, left_value) -> bool:
        return self.left.eval(left_value) and self.right.eval(left_value)

    def prefix_notation(self, trait: Trait = None, traitable_class: BTraitableClass = None) -> dict:
        res = self.left.prefix_notation(trait, traitable_class)
        res.update(self.right.prefix_notation(trait, traitable_class))
        return res


class BoolOp(Op, ABC, label=''):
    s_false: IN = IN([])

    @classmethod
    def _simplify(cls, expressions: tuple, false: IN) -> list: ...

    def __new__(cls, *expressions):
        expressions = cls._simplify(expressions, cls.s_false)
        if len(expressions) == 1:
            return expressions[0]

        obj = super().__new__(cls, expressions)
        return obj

    def prefix_notation(self, trait: Trait = None, traitable_class: BTraitableClass = None) -> dict:
        rvalues = [pn for e in self.right_value if (pn := e.prefix_notation(trait, traitable_class))]
        return {self.label: rvalues} if rvalues else {}


class AND(BoolOp):
    @classmethod
    def _simplify(cls, expressions, false):
        return [false] if false in expressions else expressions

    def eval(self, left_value) -> bool:
        return all(e.eval(left_value) for e in self.right_value)


class OR(BoolOp):
    @classmethod
    def _simplify(cls, expressions, false):
        expressions = [expression for expression in expressions if expression is not false]
        return [false] if not expressions else expressions

    def eval(self, left_value) -> bool:
        return any(e.eval(left_value) for e in self.right_value)


class f(_filter):
    def __init__(self, _f: _filter = None, _t: BTraitableClass = None, **named_expressions):
        self.filter = _f
        self.traitable_class = _t
        self.named_expressions = {
            name: expression if isinstance(expression, _filter) else EQ(expression) for name, expression in named_expressions.items()
        }

    def eval(self, traitable_or_dict) -> bool:
        if self.filter:
            if not self.filter.eval(traitable_or_dict):
                return False

        return all(item.eval(traitable_or_dict[name]) for name, item in self.named_expressions.items())

    def prefix_notation(self, trait: Trait = None, traitable_class: BTraitableClass = None) -> dict:
        traitable_class = traitable_class or self.traitable_class
        trait_dir = traitable_class.trait_dir() if traitable_class else {}
        clause = {name: pn for name, item in self.named_expressions.items() if (pn := item.prefix_notation(trait_dir.get(name), traitable_class))}
        if self.filter:
            filter_clause = self.filter.prefix_notation(traitable_class=traitable_class)
            clause = {AND.label: [filter_clause, clause]} if clause else filter_clause

        return clause



from __future__ import annotations

import functools
import operator
import sys
import warnings
from abc import ABC, abstractmethod
from contextlib import nullcontext
from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, Any, get_origin

from py10x_kernel import BTraitable, BTraitableClass, BTraitableProcessor, BTraitFlags, OsUser, XCache
from typing_extensions import Self, deprecated

import core_10x.concrete_traits as concrete_traits
from core_10x.environment_variables import EnvVars
from core_10x.global_cache import cache
from core_10x.nucleus import Nucleus

from core_10x.package_refactoring import PackageRefactoring
from core_10x.py_class import PyClass
from core_10x.rc import RC, RC_TRUE
from core_10x.resource import Resource
from core_10x.trait import TRAIT_METHOD, BoundTrait, ClassTrait, T, Trait, trait_value
from core_10x.trait_definition import (
    RT,
    M,
    TraitDefinition,
    TraitModification,
    Ui,
)
from core_10x.trait_filter import LE, f
from core_10x.traitable_id import ID
from core_10x.ts_store import TS_STORE, TsStore
from core_10x.sec_keys import SecKeys
from core_10x.xnone import XNone, XNoneType

if TYPE_CHECKING:
    from collections.abc import Generator

    from core_10x.ts_store import TsCollection


class TraitAccessor:
    __slots__ = ('cls', 'obj')

    def __init__(self, obj: Traitable, cls = None):
        if obj:
            self.cls = obj.__class__
            self.obj = obj
        else:
            self.cls = cls
            self.obj = None

    def __getattr__(self, trait_name: str):
        cls = self.cls
        trait = cls.trait(trait_name, throw = True)
        obj = self.obj
        return BoundTrait(obj, trait) if obj else ClassTrait(cls, trait)

    def __call__(self, trait_name: str, throw=True) -> Trait:
        return self.cls.trait(trait_name, throw=True)


class UnboundTraitAccessor:
    __slots__ = ()

    def __get__(self, instance, owner):
        return TraitAccessor(instance) if instance else TraitAccessor(None, cls = owner)


COLL_NAME_TAG = '_collection_name'


class StorageHelperDescriptor:
    """Resolves s_storage_helper via __mro__ when in AsOf context; otherwise returns real helper."""

    def __get__(self, obj: Traitable | None, owner: type[Traitable] | None = None) -> AbstractStorableHelper | Self:
        if owner is None:
            return self

        if not (helper := owner.s_storage_helper_cached):
            helper_as_of: StorableHelperAsOf = (
                next(
                    (
                        base_helper
                        for base in owner.__mro__
                        if issubclass(base, Traitable) and isinstance(base_helper := base.s_storage_helper_cached, StorableHelperAsOf)
                    ),
                    None,
                )
                if owner.s_history_class
                else None
            )
            if helper_as_of:
                helper = StorableHelperAsOf(owner, helper_as_of.as_of_time)
            elif not owner.is_storable():
                helper = NotStorableHelper(owner)
            else:
                helper = StorableHelperWithHistory(owner) if owner.s_history_class else StorableHelper(owner)

            owner.s_storage_helper_cached = helper

        return helper


class TraitableMetaclass(type(BTraitable)):
    def __new__(cls, name, bases, class_dict, **kwargs):
        if '__init__' in class_dict and name != 'Traitable':
            raise TypeError(f'Overriding __init__ is not allowed in {name}. Use __post_init__ instead.')
        return super().__new__(cls, name, bases, class_dict | {'__slots__': ()}, **kwargs)


class Traitable(BTraitable, Nucleus, metaclass=TraitableMetaclass):
    s_dir = {}
    s_default_trait_factory = RT
    s_own_trait_definitions = None
    T = UnboundTraitAccessor()

    _collection_name: str = XNone
    _rev: int = XNone

    @classmethod
    @cache
    def rev_trait(cls) -> Trait:
        trait_name = Nucleus.REVISION_TAG()
        return Trait.create(
            trait_name,
            T(0, T.RESERVED, data_type=int),
        )

    @classmethod
    @cache
    def collection_name_trait(cls) -> Trait:
        return Trait.create(
            COLL_NAME_TAG,
            T(T.RESERVED | T.RUNTIME, data_type=str),
        )

    def _collection_name_get(self) -> str:
        return self.id().collection_name or XNone

    def _collection_name_set(self, trait, value) -> RC:
        self.id().collection_name = value
        return RC_TRUE

    @classmethod
    def own_trait_definitions(cls) -> Generator[tuple[str, TraitDefinition]]:
        class_dict = dict(cls.__dict__)
        own_trait_definitions = class_dict.get('s_own_trait_definitions')
        if own_trait_definitions is not None:
            yield from cls.s_own_trait_definitions.items()
            return

        module_dict = sys.modules[cls.__module__].__dict__ if cls.__module__ else {}
        type_annotations = class_dict.get('__annotations__',{})
        type_annotations |= {k: XNoneType for k, v in class_dict.items() if isinstance(v, TraitModification) and k not in type_annotations}

        rc = RC(True)
        for trait_name, trait_def in class_dict.items():
            if isinstance(trait_def, TraitDefinition) and trait_name not in type_annotations:
                rc <<= f'{trait_name} = T(...), but the trait is missing a data type annotation. Use `Any` if needed.'

        for trait_name, dt in type_annotations.items():
            trait_def = class_dict.get(trait_name, class_dict)
            if trait_def is not class_dict and not isinstance(trait_def, TraitDefinition):
                continue

            old_trait: Trait = cls.s_dir.get(trait_name)
            if not (dt := cls.check_trait_type(trait_name, trait_def, old_trait, dt, class_dict, module_dict, rc)):
                continue

            if trait_def is class_dict:  # -- only annotation, not in class_dict
                trait_def = cls.s_default_trait_factory()

            if isinstance(trait_def, TraitModification):
                if not old_trait:
                    rc <<= f'{trait_name} = M(...), but the trait is not defined previously'
                    continue
                trait_def = trait_def.apply(old_trait.t_def)
                if dt is not XNoneType:  # -- the data type is also being modified
                    trait_def.data_type = dt
            else:
                trait_def.data_type = dt

            yield trait_name, trait_def

        rc.throw(TypeError)

    @classmethod
    def check_trait_type(cls, trait_name, trait_def, old_trait, dt, class_dict, module_dict, rc):
        if not dt and old_trait:
            return old_trait.data_type

        if isinstance(dt, str):
            try:
                dt = eval(dt, module_dict, class_dict)
            except Exception as e:
                rc <<= f'Failed to evaluate type annotation string `{dt}` for `{trait_name}`: {e} in {module_dict["__name__"]}'
                return None
        if dt is Self:
            dt = cls
        if dt is Any:
            dt = XNoneType
        try:
            dt_valid = isinstance(dt, type) or get_origin(dt) is not None
        except TypeError:
            dt_valid = False
        if not dt_valid:
            rc <<= f'Expected type for `{trait_name}`, but found {dt} of type {type(dt)}.'
            return None

        if trait_def is class_dict:  # -- only annotation, not in class_dict
            if old_trait and dt is not old_trait.data_type:
                rc <<= f'Attempt to implicitly redefine type for previously defined trait `{trait_name}`.  Use M() if needed.'
                return None

        return dt

    @classmethod
    def inherited_trait_dirs(cls) -> Generator[dict[str, Trait]]:
        return (base.s_dir for base in reversed(cls.__bases__) if issubclass(base, Traitable))

    @classmethod
    def _post_build_trait_dir(cls, trait_dir: dict):
        pass

    @classmethod
    def build_trait_dir(cls):
        class_dict = dict(cls.__dict__)
        module_dict = sys.modules[cls.__module__].__dict__ if cls.__module__ else {}
        type_annotations = class_dict.get('__annotations__',{})
        trait_dir = cls.s_dir
        reserved_storable_traits = {
            Nucleus.REVISION_TAG(): cls.rev_trait(),
            COLL_NAME_TAG: cls.collection_name_trait(),
        }
        trait_dir |= reserved_storable_traits
        trait_dir |= functools.reduce(operator.or_, cls.inherited_trait_dirs(), {})

        rc = RC(True)
        for trait_name, old_trait in trait_dir.items():
            trait_def = class_dict.get(trait_name, class_dict)
            if trait_def is class_dict and any(func_name in class_dict for func_name in Trait.method_defs(trait_name)):
                if cls.check_trait_type(trait_name, trait_def, old_trait, type_annotations.get(trait_name), class_dict, module_dict, rc):
                    trait_def = old_trait.t_def.copy()
                    trait_dir[trait_name] = Trait.create(trait_name, trait_def)

        for trait_name, trait_def in cls.own_trait_definitions():
            trait_def.name = trait_name
            trait_dir[trait_name] = Trait.create(trait_name, trait_def)

        cls._post_build_trait_dir(trait_dir)

        if not cls.is_storable():
            for trait_name in reserved_storable_traits:
                del trait_dir[trait_name]
        rc.throw(TypeError)

    @classmethod
    def traits(cls, flags_on: int | BTraitFlags = 0, flags_off: int | BTraitFlags = 0) -> Generator[Trait, None, None]:
        return (t for t in cls.s_dir.values() if (not flags_on or t.flags_on(flags_on)) and not t.flags_on(flags_off))

    def __hash__(self):
        return hash(self.id())

    @classmethod
    def trait(cls, trait_name: str, throw: bool = False) -> Trait:
        trait = cls.s_dir.get(trait_name)
        if trait is None and throw:
            raise TypeError(f'{cls} - unknown trait {trait_name}')

        return trait

    @classmethod
    def is_id_endogenous(cls) -> bool:
        return cls.s_bclass.is_id_endogenous()

    @classmethod
    def is_storable(cls) -> bool:
        return cls.s_bclass.is_storable()

    # @staticmethod
    # def find_storable_class(class_id: str):
    #     traitable_class = PackageRefactoring.find_class(class_id)
    #     if not issubclass(traitable_class, Traitable) or not traitable_class.is_storable():
    #         raise TypeError(f'{traitable_class} is not a storable Traitable')
    #
    #     return traitable_class

    s_bclass: BTraitableClass = None
    s_embeddable        = False
    s_custom_collection = False
    s_history_class = XNone  # -- will be set in __init__subclass__ for storable traitables unless keep_history = False. affects storage only.
    s_immutable = (
        XNone  # -- will be turned on in __init__subclass__ for storable traitables without history unless immutable=False. affects storage only.
    )
    s_direct_subclasses: list[type[Traitable]] = []
    s_storage_helper: AbstractStorableHelper = StorageHelperDescriptor()
    s_storage_helper_cached: AbstractStorableHelper | None = None

    def __init_subclass__(
        cls,
        root_class              = False,    #-- the class is a 'root' with no intent to instantiate - skip irrelevant checks!
        embeddable: bool        = None,     #-- if instances of cls may be embedded in other traitables
        custom_collection: bool = None,     #-- if instance(s) of cls may work with a specific collection
        keep_history: bool      = None,     #-- if revisions are kept in store
        immutable: bool         = None,     #-- if instances in store in are immutable
        **kwargs
    ):
        super().__init_subclass__(**kwargs)     #-- for cooperative (possible) multiple inheritance

        if embeddable is not None:
            cls.s_embeddable = embeddable

        if custom_collection is not None:
            cls.s_custom_collection = custom_collection

        cls.s_direct_subclasses = []
        for base in cls.__bases__:
            if issubclass(base, Traitable):
                base.s_direct_subclasses.append(cls)

        cls.s_dir = {}
        cls.s_bclass = BTraitableClass(cls)

        cls.build_trait_dir()  # -- build cls.s_dir from trait definitions in cls.__dict__

        if cls.is_storable():
            if keep_history is False:
                cls.s_history_class = None
            if cls.s_history_class is not None:
                cls.s_history_class = TraitableHistory.history_class(cls)

            cls.s_immutable = cls.s_history_class is None if immutable is None else immutable
        else:
            cls.s_history_class = XNone

        if cls.s_embeddable:
            cls.collection = cls._embedded_collection

        rc = RC(True)
        for trait_name, trait in cls.s_dir.items():
            trait.set_trait_funcs(cls, rc)
            trait.check_integrity(cls, rc)
            setattr(cls, trait_name, trait)
        cls.check_integrity(root_class, rc)
        rc.throw()

    @classmethod
    def _embedded_collection(cls, _coll_name):
        raise AssertionError(f"{cls} - 'embeddable' traitable may not have a collection")

    @classmethod
    def check_integrity(cls, root_class: bool, rc: RC):
        if cls.s_embeddable:
            if cls.is_id_endogenous():
                rc.add_error(f"{cls} is 'embeddable' and may NOT have ID traits")
            elif not root_class and not cls.is_storable():
                rc.add_error(f"{cls} is 'embeddable', but all its traits declared RUNTIME")

        if cls.is_storable() and (rt_id_trait := next((trait for trait in cls.traits(flags_on=T.ID) if trait.flags_on((T.RUNTIME))), None)):
            rc.add_error(f'{cls}.{rt_id_trait.name} is a RUNTIME ID trait - traitable must not be storable (all traits must be RUNTIME)')

    def __init__(self, _id: ID = None, _collection_name: str = None, _skip_init=False, _replace=False, **trait_values):
        cls = self.__class__

        if _id is not None:
            assert _collection_name is None, f'{self.__class__}(id_value) may not be invoked with _collection_name'
            assert not trait_values, f'{self.__class__}(id_value) may not be invoked with trait_values'
            super().__init__(cls.s_bclass, _id)
        else:
            super().__init__(cls.s_bclass, ID(collection_name=_collection_name))
            if not _skip_init:
                self.initialize(trait_values, _replace=_replace)

        self.__post_init__()

    def __post_init__(self):
        # Default no-op; users can override this in subclasses
        pass

    @classmethod
    def existing_instance(cls, _collection_name: str = None, _throw: bool = True, **trait_values) -> Traitable | None:
        if not cls.is_storable() and cls.is_id_endogenous():  # runtime endogenous instances are created on the fly
            return cls(_collection_name=_collection_name, **trait_values)

        obj = cls(_collection_name=_collection_name, _skip_init=True)
        if not obj.accept_existing(trait_values):
            if _throw:
                raise ValueError(f'Instance does not exist: {cls}({trait_values})')
            return None

        return obj

    @classmethod
    def existing_instance_by_id(cls, _id: ID = None, _id_value: str = None, _collection_name: str = None, _throw: bool = True) -> Traitable | None:
        if _id is None:
            _id = ID(_id_value, _collection_name)
        obj = cls(_id=_id)
        if obj.id_exists():
            return obj

        if _throw:
            raise ValueError(f'Instance does not exist: {cls}.{_id_value}')

        return None

    @classmethod
    def existing_instances_by_filter(cls, query: f, _coll_name: str = None, _at_most = 0, _order: dict = None) -> list[Traitable]:
        # ids = cls.load_ids(query = query, _coll_name = _coll_name, _at_most = _at_most, _order = _order)
        # return [ obj for id in ids if(obj := cls.existing_instance_by_id(_id = id, _collection_name = _coll_name, _throw = False)) ]
        ids_in_store = cls.load_ids(query = query, _coll_name = _coll_name, _at_most = _at_most, _order = _order)
        cache = BTraitableProcessor.current().cache()
        ids_in_memory = cache.object_ids_by_class(cls.s_bclass)
        ids_sought = { id for id in ids_in_memory if query.eval(cls(_id = id)) }
        ids_sought.update(ids_in_store)
        return [ cls(_id = id) for id in ids_sought ]

    @classmethod
    @deprecated('Use create_or_replace instead.')
    def update(cls, **kwargs) -> Traitable:
        return cls(**kwargs, _replace=True)

    @classmethod
    def new_or_replace(cls, **kwargs) -> Traitable:
        return cls(**kwargs, _replace=True)

    @classmethod
    @cache
    def runtime(cls) -> type[Traitable]:
        class Runtime(cls):
            @classmethod
            def _post_build_trait_dir(cls, trait_dir: dict):
                for trait in trait_dir.values():
                    trait.set_flags(T.RUNTIME.value())
        Runtime._post_build_trait_dir = cls._post_build_trait_dir
        return Runtime

    def set_values(self, _ignore_unknown_traits = True, **trait_values) -> RC:
        return self._set_values(trait_values, _ignore_unknown_traits)

    def __getitem__(self, item):
        return self.get_value(item)

    def __setitem__(self, key, value):
        return self.set_value(key, value)

    #===================================================================================================================
    #   The following methods are available from c++
    #
    #   get_value(trait-or-name, *args) -> Any
    #   set_value(trait-or_name, value: Any, *args) -> RC
    #   raw_value(trait-or_name, value: Any, *args) -> RC
    #   invalidate_value(trait-or-name)
    #   verify_trait(trait) -> RC               #-- semantically equiv to verify_trait_value(trait. get_value(trait))
    #   verify_trait_value(trait, value) -> RC
    #   verify()                                #-- verifies each trait and calls post_verify()
    #===================================================================================================================

    #===================================================================================================================
    #   Nucleus related methods
    #===================================================================================================================

    def serialize(self, embed: bool):
        return self.serialize_nx(embed)

    @classmethod
    def is_bundle(cls) -> bool:
        return cls.serialize_class_id is not Traitable.serialize_class_id

    @classmethod
    def serialize_class_id(cls) -> str | None:
        return None

    @classmethod
    def deserialize_class_id(cls, serialized_class_id: str):
        return cls

    @classmethod
    def deserialize(cls, serialized_data) -> Traitable:
        return Traitable.deserialize_nx(cls.s_bclass, serialized_data)

    def to_str(self) -> str:
        return f'{self.id()}'

    @classmethod
    def from_str(cls, s: str) -> Nucleus:
        # return cls(ID(s))  # collection name?
        return cls.existing_instance_by_id(_id_value=s)

    @classmethod
    def from_any_xstr(cls, value) -> Nucleus:
        if isinstance(value, dict):
            return cls(**value)

        raise TypeError(f'{cls}.from_any_xstr() expects a dict, got {value})')

    @classmethod
    def same_values(cls, value1, value2) -> bool:
        if type(value2) is str:
            return value1.id().value == value2

        return value1.id() == value2.id()

    #===================================================================================================================
    #   Storage related methods
    #===================================================================================================================

    @classmethod
    def store(cls) -> TsStore:
        store: TsStore = TS_STORE.current_resource()    #-- if current TsStore is set, use it!
        if not store:
            store = cls.store_per_class()               #-- otherwise, use per class store, if any

        if not store:
            raise OSError(f'{cls} - no Traitable Store is available')

        return store

    @staticmethod
    @cache
    def vault_store() -> TsStore:
        uri = SecKeys.check_vault_uri()
        spec = TsStore.spec_from_uri(uri)
        store_class: type[TsStore] = spec.resource_class
        is_running, with_auth = store_class.is_running_with_auth(spec.hostname())
        if not is_running:
            raise OSError(f"Vault Store '{uri}' is not running")

        if with_auth:
            pwd = SecKeys.retrieve_vault_password(vault_uri = uri)
            spec.set_credentials(pwd)

        return store_class.instance(**spec.kwargs)

    @staticmethod
    def store_from_uri(uri: str) -> TsStore:
        spec = TsStore.spec_from_uri(uri)
        store_class: type[ TsStore ] = spec.resource_class
        is_running, with_auth = store_class.is_running_with_auth(spec.hostname())
        if not is_running:
            raise OSError(f"TsStore '{uri}' is not running")

        if with_auth:
            with Traitable.vault_store():
                ra = VaultResourceAccessor.retrieve_ra(uri)
                return ra.resource

        return store_class.instance(**spec.kwargs)

    @staticmethod
    @cache
    def main_store() -> TsStore:
        v_store_uri = EnvVars.var.main_ts_store_uri
        if not v_store_uri:
            raise OSError(f'No Traitable Store is specified: neither explicitly, nor via environment variable {EnvVars.var_name(v_store_uri)}')

        return Traitable.store_from_uri(v_store_uri.value)

    @classmethod
    @cache
    def store_per_class(cls) -> TsStore:
        store = Traitable.main_store()
        #-- check if there's a specific store association with this cls
        if EnvVars.use_ts_store_per_class:
            with store:
                ts_uri = TsClassAssociation.ts_uri(cls)

            if ts_uri:
                store = Traitable.store_from_uri(ts_uri)

        return store

    @classmethod
    def collection(cls, _coll_name: str = None) -> TsCollection | None:
        return cls.s_storage_helper.collection(_coll_name)

    @classmethod
    def exists_in_store(cls, id: ID) -> bool:
        return cls.s_storage_helper.exists_in_store(id)

    @classmethod
    def load_data(cls, id: ID) -> dict | None:
        return cls.s_storage_helper.load_data(id)

    @classmethod
    def delete_in_store(cls, id: ID) -> RC:
        return cls.s_storage_helper.delete_in_store(id)

    @classmethod
    def load(cls, id: ID) -> Traitable | None:
        return cls.s_storage_helper.load(id)

    @classmethod
    def load_many(cls, query: f = None, _coll_name: str = None, _at_most: int = 0, _order: dict = None, _deserialize=True) -> list[Self]:
        return cls.s_storage_helper.load_many(query, _coll_name, _at_most, _order, _deserialize)

    @classmethod
    def load_ids(cls, query: f = None, _coll_name: str = None, _at_most: int = 0, _order: dict = None) -> list[ID]:
        return cls.s_storage_helper.load_ids(query, _coll_name, _at_most, _order)

    @classmethod
    def delete_collection(cls, _coll_name: str = None) -> bool:
        return cls.s_storage_helper.delete_collection(_coll_name)

    def save(self, save_references=False) -> RC:
        return self.__class__.s_storage_helper.save(self, save_references=save_references)

    def delete(self) -> RC:
        return self.__class__.s_storage_helper.delete(self)

    def post_verify(self) -> RC:
        return RC_TRUE

    @classmethod
    def as_of(cls, traitable_id: ID, as_of_time: datetime) -> Self:
        return cls.s_storage_helper.as_of(traitable_id,as_of_time)

    @classmethod
    def history(cls, _at_most: int = 0, _filter: f = None, _deserialize=False, _collection_name: str = None, _before: datetime = None, **named_filters) -> list:
        """Get history entries for this traitable class."""
        return cls.s_storage_helper.history(_at_most,_filter,_deserialize,_collection_name,_before,**named_filters)

    @classmethod
    def latest_revision(cls, traitable_id: ID, timestamp: datetime = None, deserialize: bool = False) -> dict | TraitableHistory | None:
        """Get the latest revision of a traitable from history."""
        return cls.s_storage_helper.latest_revision(traitable_id,timestamp,deserialize)

    @classmethod
    def restore(cls, traitable_id, timestamp: datetime = None, save=False) -> bool:
        """Restore a traitable to a specific point in time."""
        return cls.s_storage_helper.restore(traitable_id,timestamp,save)

Traitable.s_bclass = BTraitableClass(Traitable)

@dataclass
class AbstractStorableHelper(ABC):
    traitable_class: type[Traitable]

    @abstractmethod
    def collection(self, _coll_name: str = None) -> TsCollection | None: ...

    @abstractmethod
    def exists_in_store(self, id: ID) -> bool: ...

    @abstractmethod
    def load_data(self, id: ID) -> dict | None: ...

    @abstractmethod
    def delete_in_store(self, id: ID) -> RC: ...

    @abstractmethod
    def load(self, id: ID) -> Traitable | None: ...

    @abstractmethod
    def load_many(self, query: f = None, _coll_name: str = None, _at_most: int = 0, _order: dict = None, _deserialize=True) -> list[Traitable]: ...

    @abstractmethod
    def load_ids(self, query: f = None, _coll_name: str = None, _at_most: int = 0, _order: dict = None) -> list[ID]: ...

    @abstractmethod
    def delete_collection(self, _coll_name: str = None) -> bool: ...

    @abstractmethod
    def save(self, traitable: Traitable, save_references: bool) -> RC: ...

    @abstractmethod
    def _save_serialized(self, coll, serialized_data, old_rev): ...

    @abstractmethod
    def delete(self, traitable: Traitable) -> RC: ...

    @abstractmethod
    def as_of(self, traitable_id: ID, as_of_time: datetime) -> Traitable: ...

    @abstractmethod
    def history(self, _at_most: int = 0, _filter: f = None, _deserialize=False, _collection_name: str = None, _before: datetime = None, **named_filters) -> list: ...

    @abstractmethod
    def latest_revision(self, traitable_id: ID, timestamp: datetime = None, deserialize: bool = False) -> dict | TraitableHistory | None: ...

    @abstractmethod
    def restore(self, traitable_id, timestamp: datetime = None, save=False) -> bool: ...


class NotStorableHelper(AbstractStorableHelper):
    def collection(self, _coll_name: str = None) -> TsCollection | None:
        return None

    def exists_in_store(self, id: ID) -> bool:
        return False

    def load_data(self, id: ID) -> dict | None:
        return None

    def delete_in_store(self, id: ID) -> RC:
        return RC(False, f'{self.traitable_class} is not storable')

    def load(self, id: ID) -> Traitable | None:
        return None

    def load_many(self, query: f = None, _coll_name: str = None, _at_most: int = 0, _order: dict = None, _deserialize=True) -> list[Traitable]:
        return []

    def load_ids(self, query: f = None, _coll_name: str = None, _at_most: int = 0, _order: dict = None) -> list[ID]:
        return []

    def delete_collection(self, _coll_name: str = None) -> bool:
        return False

    def save(self, traitable: Traitable, save_references: bool) -> RC:
        return RC(False, f'{self.traitable_class} is not storable')

    def _save_serialized(self, coll, serialized_data, old_rev):
        return old_rev

    def delete(self, traitable: Traitable) -> RC:
        return RC(False, f'{self.traitable_class} is not storable')

    def as_of(self, traitable_id: ID, as_of_time: datetime) -> Traitable:
        raise RuntimeError(f'{self.traitable_class} is not storable')

    def history(self, _at_most: int = 0, _filter: f = None, _deserialize=False, _collection_name: str = None, _before: datetime = None, **named_filters) -> list:
        raise RuntimeError(f'{self.traitable_class} is not storable')

    def latest_revision(self, traitable_id: ID, timestamp: datetime = None, deserialize: bool = False) -> dict | TraitableHistory | None:
        raise RuntimeError(f'{self.traitable_class} is not storable')

    def restore(self, traitable_id, timestamp: datetime = None, save=False) -> bool:
        raise RuntimeError(f'{self.traitable_class} is not storable')



class StorableHelper(AbstractStorableHelper):
    def collection(self, _coll_name: str = None) -> TsCollection:
        cls = self.traitable_class
        cname = _coll_name or PackageRefactoring.find_class_id(cls)
        return cls.store().collection(cname)

    def exists_in_store(self, id: ID) -> bool:
        return self.traitable_class.collection(_coll_name=id.collection_name).id_exists(id.value)

    def load_data(self, id: ID) -> dict | None:
        return self.traitable_class.collection(_coll_name=id.collection_name).load(id.value)

    def delete_in_store(self, id: ID) -> RC:
        cls = self.traitable_class
        coll = cls.collection(_coll_name=id.collection_name)
        if not coll:
            return RC(False, f'{cls} - no store available')
        if not coll.delete(id.value):
            return RC(False, f'{cls} - failed to delete {id.value} from {coll}')
        return RC_TRUE

    def load(self, id: ID) -> Traitable | None:
        return self.traitable_class.s_bclass.load(id)

    def _find(self, query: f = None, _coll_name: str = None, _at_most: int = 0, _order: dict = None):
        # TODO FUTURE - current state as history query?
        cls = self.traitable_class
        coll = cls.collection(_coll_name=_coll_name)
        return coll.find(f(query, cls.s_bclass), _at_most=_at_most, _order=_order)

    def load_many(
        self, query: f = None, _coll_name: str = None, _at_most: int = 0, _order: dict = None, _deserialize: bool = True
    ) -> list[Traitable] | list[dict]:
        cursor = self._find(query=query, _coll_name=_coll_name, _at_most=_at_most, _order=_order)

        if not _deserialize:
            return list(cursor)

        f_deserialize = functools.partial(Traitable.deserialize_object, self.traitable_class.s_bclass, _coll_name)
        return [f_deserialize(serialized_data) for serialized_data in cursor]

    def load_ids(self, query: f = None, _coll_name: str = None, _at_most: int = 0, _order: dict = None) -> list[ID]:
        id_tag = self.traitable_class.collection(_coll_name=_coll_name).s_id_tag  # better?
        cursor = self._find(query=query, _coll_name=_coll_name, _at_most=_at_most, _order=_order)
        return [ID(serialized_data.get(id_tag), _coll_name) for serialized_data in cursor]

    def delete_collection(self, _coll_name: str = None) -> bool:
        cls = self.traitable_class
        store = cls.store()
        if not store:
            return False
        cname = _coll_name or PackageRefactoring.find_class_id(cls)
        return store.delete_collection(collection_name=cname)

    def save(self, traitable: Traitable, save_references: bool) -> RC:
        rc = traitable.verify()
        if not rc:
            return rc

        rc = traitable.share(False)  # -- not accepting existing traitable values, if any
        if not rc:
            return rc


        try:
            serialized_data = traitable.serialize_object(save_references)

            if not serialized_data:  # -- it's a lazy instance - no reason to load and re-save
                return RC_TRUE

            coll = self.traitable_class.collection(traitable.id().collection_name)
            if not coll:
                return RC(False, f'{self.__class__} - no store available')

            with self._transaction_ctx():
                rev = self._save_serialized(coll, serialized_data, traitable.get_revision())
        except Exception as e:
            return RC(False, f'Error saving traitable: {e}')

        traitable.set_revision(rev)
        return RC_TRUE

    def _transaction_ctx(self):
        return nullcontext()

    def _save_serialized(self, coll, serialized_data, old_rev) -> int:
        return coll.save_new(serialized_data) if self.traitable_class.s_immutable else coll.save(serialized_data)

    def delete(self, traitable: Traitable) -> RC:
        rc = self.delete_in_store(traitable.id())
        if rc:
            traitable.set_revision(0)
        return rc

    def as_of(self, traitable_id: ID, as_of_time: datetime) -> Traitable:
        raise RuntimeError(f'{self.traitable_class} does not keep history')

    def history(self, _at_most: int = 0, _filter: f = None, _deserialize=False, _collection_name: str = None, _before: datetime = None, **named_filters) -> list:
        raise RuntimeError(f'{self.traitable_class} does not keep history')

    def latest_revision(self, traitable_id: ID, timestamp: datetime = None, deserialize: bool = False) -> dict | TraitableHistory | None:
        raise RuntimeError(f'{self.traitable_class} does not keep history')

    def restore(self, traitable_id, timestamp: datetime = None, save=False) -> bool:
        raise RuntimeError(f'{self.traitable_class} does not keep history')


class StorableHelperWithHistory(StorableHelper):
    def _transaction_ctx(self):
        return self.traitable_class.store().transaction() if EnvVars.use_ts_store_transactions else nullcontext()

    def _save_serialized(self, coll, serialized_data, old_rev) -> int:
        rev = super()._save_serialized(coll, serialized_data, old_rev)
        if rev > old_rev:
            self.traitable_class.s_history_class(
                serialized_traitable=serialized_data,
                _traitable_rev=rev,
                _collection_name=(coll.collection_name() + '#history') if self.traitable_class.s_custom_collection else XNone,
            ).save().throw()
        return rev

    def as_of(self, traitable_id: ID, as_of_time: datetime) -> Self:
        history_entry = self.traitable_class.latest_revision(traitable_id, as_of_time, deserialize=True)
        return history_entry.traitable if history_entry else None

    def history(
        self, _at_most: int = 0, _filter: f = None, _deserialize=False, _collection_name: str = None, _before: datetime = None, **named_filters
    ) -> list:
        cls = self.traitable_class

        if cls.s_custom_collection and not _collection_name:
            raise RuntimeError(f'{cls} requires custom _collection_name')

        if not cls.s_custom_collection and _collection_name:
            raise RuntimeError(f'{cls} does not support custom _collection_name')

        as_of = {'_at': LE(_before)} if _before else {}
        cursor = cls.s_history_class.load_many(
            f(_filter, **named_filters, **as_of),
            _order=dict(_traitable_id=1, _at=-1, _traitable_rev=-1),
            _at_most=_at_most,
            _deserialize=_deserialize,
            _coll_name=_collection_name + '#history' if _collection_name else None,
        )

        return list(cursor)

    def latest_revision(self, traitable_id: ID, timestamp: datetime = None, deserialize: bool = False) -> dict | TraitableHistory | None:
        """Get the latest revision of a traitable from history."""
        for entry in self.traitable_class.history(
            _filter=f(_traitable_id=traitable_id.value),
            _collection_name=traitable_id.collection_name,
            _at_most=1,
            _deserialize=deserialize,
            _before=timestamp,
        ):
            return entry  # Return the raw history entry dict

        return None

    def restore(self, traitable_id, timestamp: datetime = None, save=False) -> bool:
        """Restore a traitable to a specific point in time."""
        history_entry = self.traitable_class.latest_revision(
            traitable_id,
            timestamp,
            deserialize=True,
        )
        if not history_entry or not history_entry.traitable:
            return False

        if save:
            return bool(
                self.traitable_class.collection(
                    traitable_id.collection_name,
                ).save_new(
                    {'$set': history_entry.serialized_traitable},
                    overwrite=True,
                )
            )
        return True


@dataclass
class StorableHelperAsOf(StorableHelperWithHistory):
    as_of_time: datetime

    def _find(self, query: f = None, _coll_name: str = None, _at_most: int = 0, _order: dict = None):
        last_id = None
        for item in self.traitable_class.history(
            _filter=query,
            _before=self.as_of_time,
            _collection_name=_coll_name,
            _deserialize=True,
        ):
            # TODO: optimize by returning one entry per id using a pipeline query
            if last_id != item._traitable_id:
                last_id = item._traitable_id
                yield item.serialized_traitable

    def exists_in_store(self, id: ID) -> bool:
        history_entry = self.traitable_class.latest_revision(id, self.as_of_time, deserialize=True)
        return bool(history_entry)  # TODO: optimize by not downloading the latest revision

    def load_data(self, id: ID) -> dict | None:
        history_entry = self.traitable_class.latest_revision(id, self.as_of_time, deserialize=True)
        return history_entry.serialized_traitable if history_entry else None


def __getattr__(name):
    if name == 'THIS_CLASS':  # -- to use for traits with the same Traitable class type
        warnings.warn('THIS_CLASS is deprecated; use typing.Self instead', DeprecationWarning, stacklevel=2)
        return Self
    raise AttributeError(name)


class TraitableHistory(Traitable, keep_history=False):
    s_traitable_class = None
    s_trait_name_map = dict(_traitable_id='_id', _traitable_rev='_rev')

    # fmt: off
    traitable: Traitable        = RT() // 'original traitable'
    serialized_traitable: dict  = RT()

    _at: datetime               = T() // 'time saved'
    _who: str                   = T() // 'authenticated user, if any'
    _traitable_id: str          = T() // 'original traitable id'
    _traitable_rev: int         = T() // 'original traitable rev'
    # fmt: on

    def _traitable_id_get(self) -> str:
        return self.serialized_traitable['_id']

    def serialize_object(self, save_references: bool = False):
        serialized_data = {**self.serialized_traitable, **super().serialize_object(save_references)}
        return self.store().populate(['_who', '_at'], serialized_data)

    def traitable_get(self):
        return Traitable.deserialize_object(
            self.s_traitable_class.s_bclass,
            self._collection_name.rsplit('#', 1)[0] or None,  # -- strip #history suffix
            self.serialized_traitable,
        )

    def deserialize_traits(self, serialized_data):
        hist_data = {trait.name: serialized_data.pop(trait.name, None) for trait in self.traits(flags_off=T.RUNTIME)}
        self.serialized_traitable = serialized_data | {v: hist_data[k] for k, v in self.s_trait_name_map.items()}
        return super().deserialize_traits(hist_data)

    @classmethod
    def store(cls):
        return cls.s_traitable_class.store()

    @classmethod
    def collection(cls, _coll_name: str = None) -> TsCollection | None:
        collection = cls.s_storage_helper.collection(_coll_name)
        collection.create_index('idx_by_traitable_id_time', [('_traitable_id', 1), ('_at', -1)])
        return collection

    @staticmethod
    @cache
    def history_class(traitable_class: type[Traitable]):
        module_dict = sys.modules[traitable_class.__module__].__dict__
        history_class_name = f'{traitable_class.__name__}#history'
        history_class = type(
            history_class_name,
            (TraitableHistory,),
            dict(
                s_traitable_class=traitable_class,
                s_custom_collection=traitable_class.s_custom_collection,
                __module__=traitable_class.__module__,
            ),
        )
        if traitable_class is module_dict.get(traitable_class.__name__):
            assert history_class_name not in module_dict
            module_dict[history_class_name] = history_class
        return history_class


@dataclass
class AsOfContext:
    """Context manager for time-based traitable loading.

    Applies to the given traitable_classes and all their subclasses (discovered via
    s_direct_subclasses). Default is [Traitable], so all storable Traitable subclasses
    use AsOf resolution. s_storage_helper is resolved dynamically via __mro__ and
    cached; the cache is invalidated on enter/exit for all relevant subclasses.
    """

    as_of_time: datetime
    traitable_classes: list[type[Traitable]] = field(default_factory=lambda: [Traitable])

    _btp: BTraitableProcessor | None = None

    def __post_init__(self):
        invalid_class = next((traitable_class for traitable_class in self.traitable_classes if not traitable_class.s_history_class), None)
        if invalid_class and invalid_class is not Traitable:  # we use Traitable as a special case to cover *all* traitables
            raise ValueError(f'{invalid_class} is not storable or does not keep history')

    def _reset_storage_helpers(self):
        visited: set[type[Traitable]] = set()
        stack = list(self.traitable_classes)
        while stack:
            traitable_class = stack.pop()
            if traitable_class in visited:
                continue
            visited.add(traitable_class)
            traitable_class.s_storage_helper_cached = None
            for sub in traitable_class.s_direct_subclasses:
                stack.append(sub)

    def __enter__(self):
        self._reset_storage_helpers()
        for traitable_class in self.traitable_classes:
            traitable_class.s_storage_helper_cached = StorableHelperAsOf(traitable_class, self.as_of_time)

        btp = BTraitableProcessor.create_root()
        btp.begin_using()
        self._btp = btp
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._btp is not None:
            self._btp.end_using()
            self._btp = None

        self._reset_storage_helpers()

class traitable_trait(concrete_traits.nucleus_trait, data_type=Traitable, base_class=True):
    def post_ctor(self): ...

    def check_integrity(self, cls, rc: RC):
        is_runtime = self.flags_on(T.RUNTIME)
        if self.flags_on(T.EMBEDDED):
            if is_runtime:
                rc.add_error(f'{cls.__name__}.{self.name} - may NOT be both RUNTIME and EMBEDDED')
            if not self.data_type.s_embeddable:
                rc.add_error(f"{cls.__name__}.{self.name} - class {self.data_type} must be declared 'embeddable'")

    def default_value(self):
        def_value = self.default
        if def_value is XNone:
            return def_value

        if isinstance(def_value, str):  # -- id
            return self.data_type.instance_by_id(def_value)

        if isinstance(def_value, dict):  # -- trait values
            return self.data_type(**def_value)

        raise ValueError(f'{self.data_type} - may not be constructed from {def_value}')

    def from_str(self, s: str):
        return self.data_type.from_str(s)

    def from_any_xstr(self, value):
        if not isinstance(value, dict):
            return None

        return self.data_type(**value)


class AnonymousTraitable(Traitable, root_class = True, embeddable = True):
    pass

class Bundle(Traitable):
    s_bundle_base = None
    s_bundle_members: dict = None

    def __init_subclass__(cls, members_known=False, **kwargs):
        super().__init_subclass__(**kwargs)
        if not cls.s_bundle_base:
            cls.s_bundle_base = cls
            if members_known:
                cls.s_bundle_members = {}
        else:
            assert cls.is_storable(), f'{cls} is not storable'
            base = cls.s_bundle_base
            if base:
                bundle_members = base.s_bundle_members
                if bundle_members is not None:
                    bundle_members[cls.__name__] = cls

                # cls.collection_name = base.collection_name #TODO: fix
                cls.collection = base.collection
            assert cls.s_bundle_base, 'bundle base not defined'

    @classmethod
    def serialize_class_id(cls) -> str:
        if cls.s_bundle_members is None:  # -- members unknown
            return PackageRefactoring.find_class_id(cls)
        else:
            return cls.__name__

    @classmethod
    def deserialize_class_id(cls, serialized_class_id: str):
        if not serialized_class_id:
            raise ValueError('missing serialized class ID')

        if cls.s_bundle_members is None:  # -- members are not known - class_id is a real class_id
            return PackageRefactoring.find_class(serialized_class_id)

        # -- class_id is a short class name
        actual_class = cls.s_bundle_members.get(serialized_class_id)
        if not actual_class:
            raise ValueError(f'{cls}: unknown bundle member {serialized_class_id}')

        return actual_class

#=======================================================================================================================
#   Vault related stuff
#=======================================================================================================================

class VaultUser(Traitable):
    user_id: str                    = T(T.ID)   // 'OS login'
    suspended: bool                 = T(False)

    private_key_encrypted: bytes    = T()
    public_key: bytes               = T()

    sec_keys: SecKeys               = RT(T.EVAL_ONCE)

    def user_id_get(self) -> str:
        return self.__class__.myname()

    def sec_keys_get(self) -> SecKeys:
        return SecKeys(self.private_key_encrypted, self.public_key, SecKeys.retrieve_master_password())

    @classmethod
    def is_functional_account(cls, user_id: str) -> bool:
        return user_id.split('-', 1) == EnvVars.functional_account_prefix

    @classmethod
    @cache
    def myname(cls) -> str:
        return OsUser.me.name()

    @classmethod
    @cache
    def me(cls) -> VaultUser:
        return cls.existing_instance(user_id=cls.myname())


class VaultResourceAccessor(Traitable):
    username: str           = T(T.ID)
    resource_uri: str       = T(T.ID)

    login: str              = T()
    password: bytes         = T()

    last_updated: datetime  = T(T.EVAL_ONCE)

    user: VaultUser         = RT(T.EVAL_ONCE)
    resource: Resource      = RT(T.EVAL_ONCE)

    def username_get(self) -> str:
        return VaultUser.myname()

    def last_updated_get(self) -> datetime:
        return datetime.utcnow()

    def user_get(self) -> VaultUser:
        return VaultUser.existing_instance(user_id=self.username)

    def resource_get(self) -> Resource:
        return Resource.instance_from_uri(
            self.resource_uri,
            username=self.username,
            password=self.user.sec_keys.decrypt_text(self.password),
        )

    @classmethod
    def save_ra(cls, resource_uri: str, password: str, login: str = None, username: str = XNone) -> RC:
        if login is None:
            login = username

        ra = cls(username = username, resource_uri = resource_uri)
        user = ra.user
        rc = ra.set_values(
            login       = login,
            password    = user.sec_keys.encrypt_text(password),
        )
        if rc:
            rc = ra.save()

        return rc

    #-- strictly speaking this method isn't necessary as it just returning existing_instance()
    @classmethod
    def retrieve_ra(cls, resource_uri: str, username: str = XNone) -> VaultResourceAccessor:
        return cls.existing_instance(username=username, resource_uri=resource_uri)

class NamedTsStore(Traitable):
    logical_name: str   = T(T.ID)
    uri: str            = T()

class TsClassAssociation(Traitable):
    py_canonical_name: str  = T(T.ID)
    ts_logical_name: str    = T(Ui.choice('Store Name'))

    def ts_logical_name_choices(self, trait) -> tuple:
        return tuple(nts.logical_name for nts in NamedTsStore.load_many())

    @classmethod
    @cache
    def store_per_class(cls) -> TsStore:
        return Traitable.main_store()

    @classmethod
    def ts_uri(cls, traitable_class) -> str:
        #-- 1) Check TsStore association for the class itself, its module and packages
        canonical_name = PyClass.name(traitable_class)
        while True:
            association = cls.existing_instance(py_canonical_name=canonical_name, _throw=False)
            if association:
                named_store = NamedTsStore.existing_instance(logical_name=association.ts_logical_name)
                return named_store.uri

            parts = canonical_name.rsplit('.', maxsplit=1)
            name = parts[0]
            if name == canonical_name:  #-- checked all packages bottom up
                break                   #-- no asscciation for the class. module, packages

            canonical_name = name

        #-- 2) Check TsStore association for the class' parent Traitables
        parent_classes = traitable_class.__mro__
        for pclass in parent_classes[1:]:
            if issubclass(pclass, Traitable):
                canonical_name = PyClass.name(pclass)
                association = cls.existing_instance(py_canonical_name=canonical_name, _throw=False)
                if association:
                    named_store = NamedTsStore.existing_instance(logical_name=association.ts_logical_name)
                    return named_store.uri

        return ''
#=======================================================================================================================



from __future__ import annotations

from typing import TYPE_CHECKING

import rio
import ui_10x.platform_interface as i
from ui_10x.rio.component_builder import DynamicComponent, UserSessionContext, Widget
from ui_10x.rio.internals.app import App10x

if TYPE_CHECKING:
    import uvicorn


class Dialog(Widget, i.Dialog):
    __slots__ = ('_dialog', '_modal', '_parent', '_server', 'accepted', 'on_accept', 'on_reject', 'title')
    s_component_class = rio.Column
    s_forced_kwargs = {'grow_x': True, 'grow_y': True}

    def _make_kwargs(self, **kwargs):
        kwargs = super()._make_kwargs(**kwargs)
        del kwargs['align_y']
        return kwargs

    def __init__(self, parent: Widget | None = None, children=(), title=None, on_accept=None, on_reject=None, **kwargs):
        assert isinstance(parent, Widget | None)
        super().__init__(*children, **kwargs)
        self.on_accept = self._wrapper(on_accept, accept=True)
        self.on_reject = self._wrapper(on_reject)
        self.accepted = True
        self.title = title
        self._dialog = None
        self._server = None
        self._parent = parent
        self._modal = True

    def set_window_title(self, title: str):
        self.title = title

    def _wrapper(self, func, accept=False):
        func = self.callback(func) if func else None

        def wrapper(*args):
            self.accepted = accept
            if func:
                func(*args)
            self._on_close()

        return wrapper

    def reject(self):
        self._on_close()

    def done(self, result: int):
        self._on_close()
        self.accepted = bool(result)

    def _on_close(self):
        if self._dialog:
            dialog = self._dialog.result()
            dialog._root_component.session.create_task(dialog.close())
            self._dialog = None
        elif self._server:
            self._server.should_exit = True

    def _on_server_created(self, server: uvicorn.Server):
        self._server = server

    def _on_dialog_open(self, future):
        self._dialog = future

    def exec(self):
        if self.current_session():
            print('Cannot exec() in an existing event loop - using show() instead. WARNING: this only works with callbacks!')
            self.show()
            return

        assert not self._parent, 'Parent is not allowed for top level dialog'

        title = self.title or 'Dialog'
        sessions = set()

        def on_session_start(session):
            sessions.add(session)
            print('on_session_start:', len(sessions))

        def on_session_close(session):
            for component in session._weak_components_by_id.values():
                if isinstance(component, DynamicComponent):
                    component.builder.component = None
                    component.builder.subcomponent = None
            sessions.remove(session)
            print('on_session_close:', len(sessions))

        def build():
            print('build:', len(sessions))
            if len(sessions) == 1:
                component = DynamicComponent(builder=self)
                if component.session in sessions:
                    return component

            from rio.components.error_placeholder import ErrorPlaceholder

            return ErrorPlaceholder(error_summary='Only one session is allowed for `Dialog.exec`', error_details='')

        app = rio.App(
            name=title,
            build=build,
            on_session_start=on_session_start,
            on_session_close=on_session_close,
            default_attachments=[UserSessionContext(host='localhost', dbname='test')],
        )
        debug = True
        if debug:
            from rio.debug.monkeypatches import apply_monkeypatches

            apply_monkeypatches()
        App10x(app)._run_in_window(debug_mode=debug, on_server_created=self._on_server_created)
        return self.accepted

    def show(self):
        if not self.current_session():
            self.exec()
        else:
            future = self.current_session().show_custom_dialog(
                build=self,
                on_close=self._on_close,
                modal=self._modal,
                user_closable=False,
                owning_component=self._parent.component if self._parent else None,
            )
            self.current_session().create_task(future).add_done_callback(self._on_dialog_open)

    def set_window_flags(self, flags):
        raise NotImplementedError

    def set_modal(self, modal: bool):
        self._modal = modal



import getpass

from core_10x.traitable import VaultUser, VaultResourceAccessor, Resource, TsStore, RC, RC_TRUE
from core_10x.sec_keys import SecKeys

class VaultUtils:
    MIN_CHARS   = 8
    @classmethod
    def verify_new_password(cls, pwd: str, pwd2: str) -> RC:
        rc = RC(True)
        if not pwd or not pwd2:
            rc.add_error('passwords must not be empty')
        n = len(pwd)
        if n < cls.MIN_CHARS:
            rc.add_error(f'password must be at least {cls.MIN_CHARS} characters long')
        if not any(c.isalpha() for c in pwd):
            rc.add_error(f'password must have at least one letter')
        if not any(c.isupper() for c in pwd):
            rc.add_error(f'password must have at least one capital letter')
        if not any(c.isdigit() for c in pwd):
            rc.add_error(f'password must have at least one digit')
        if pwd != pwd2:
            rc.add_error(f'passwords do not match')

        if not rc:
            rc.prepend_error_header('New password cannot be used:')

        return rc

    @classmethod
    def create_master_password(cls, ovveride = False) -> RC:
        if not ovveride:
            mp = SecKeys.retrieve_master_password(throw = False)
            if mp is not None:
                return RC(False, 'MasterPassword already exists')

        mp = getpass.getpass('Please create a memorable MasterPassword: ')
        mp2= getpass.getpass('Please re-enter your MasterPassword: ')
        rc = cls.verify_new_password(mp, mp2)
        if rc:
            SecKeys.change_master_password(mp, override = True)

        return rc

    @classmethod
    def change_master_password(cls) -> RC:
        old_mp = SecKeys.retrieve_master_password(throw = False)
        if old_mp is not None:
            proceed = input('Are you sure you want to override your existing MasterPassword? [y/n]')
            if proceed.upper() != 'Y':
                return RC_TRUE

        return cls.create_master_password(ovveride = True)

    @classmethod
    def _create_user(cls, save = True) -> VaultUser:  #-- must be called inside with vault_store block
        private_key_pem, public_key_pem = SecKeys.generate_keys()
        master_password = SecKeys.retrieve_master_password()
        me = VaultUser(
            _replace                = True,
            public_key              = public_key_pem,
            private_key_encrypted   = SecKeys.encrypt_private_key(private_key_pem, master_password),
        )
        if save:
            me.save().throw()

        return me

    @classmethod
    def get_vault(cls, vault_uri: str = None) -> TsStore:
        if vault_uri is None:
            vault_uri = input('Please enter vault URI: ')

        vault_password = SecKeys.retrieve_vault_password(vault_uri = vault_uri, throw = False)
        if not vault_password:
            vault_password = getpass.getpass('Please enter password for the vault: ')

        return TsStore.instance_from_uri(vault_uri, username = VaultUser.myname(), password = vault_password)

    @classmethod
    def register_vault(cls) -> RC:
        if not SecKeys.retrieve_master_password(throw = False):
            rc = cls.create_master_password()
            if not rc:
                return rc

        vault_uri = input('Please enter vault URI (e.g., mongodb://vaultdb.xxx.io[:27017]/vault): ')
        password = getpass.getpass('Please enter password for the vault (should have been given to you by admin): ')

        username = VaultUser.myname()
        vault_db = TsStore.instance_from_uri(vault_uri, username = username, password = password)

        with vault_db:
            if not VaultUser.existing_instance(user_id = username, _throw = False):
                cls._create_user()

        SecKeys.change_vault_password(password, vault_uri = vault_uri, override = True)
        return RC_TRUE

    @classmethod
    def enter_resource_accessors(cls) -> RC:
        vault_db = cls.get_vault()
        with vault_db:
            while True:
                user_id = input('Please enter user ID the resource accessors are for: ')
                if not user_id:
                    break

                vault_user = VaultUser.existing_instance(user_id = user_id, _throw = False)
                if not vault_user:
                    print(f"User '{user_id}' has not registered yet. Please inform '{user_id}'")
                    continue

                while True:
                    uri = input('Please enter resource URI: ')
                    if not uri:
                        break

                    while True:
                        login = input(f"Please enter login name for'{user_id}': ")
                        if not login:
                            break

                        password = getpass.getpass(f"Please enter password for '{login}': ")
                        if not password:
                            continue

                        try:
                            Resource.instance_from_uri(uri, username = login, password = password)
                        except Exception as ex:
                            print(f'{str(ex)}')

                        VaultResourceAccessor.save_ra(uri, password, login = login, username = user_id)




from core_10x.code_samples.person import Person
from core_10x.traitable import NamedTsStore, T, Traitable, TsClassAssociation, TsStore


class Dummy1(Traitable):
    text: str   = T(T.ID)

class Dummy2(Traitable):
    text: str   = T(T.ID)

class Dummy3(Dummy2):
    ...

if __name__ == '__main__':
    from core_10x.environment_variables import EnvVars
    from core_10x.manual_tests.ts_class_association_test import Dummy1, Dummy2, Dummy3
    from core_10x.py_class import PyClass

    #-- 0) Create a Person instance and save - must go to the main TsStore - no association
    p = Person(first_name = 'A', last_name = 'B')
    p.save().throw()
    db = TsStore.instance_from_uri(EnvVars.main_ts_store_uri)
    with db:
        assert Person.exists_in_store(p.id()), f'Person must have been stored to {EnvVars.main_ts_store_uri}'

    #-- 1) Create NamedTsStore objects referring to all available extra TsStores for Class Associations
    extra_uris = dict(
        dummy1 = 'mongodb://localhost/dummy1',
        dummy2 = 'mongodb://localhost/dummy2',
    )
    for name, uri in extra_uris.items():
        #store = TsStore.instance_from_uri(uri)
        ns = NamedTsStore(_replace = True,
            logical_name    = name,
            uri             = uri,
        )
        ns.save().throw()

    #-- 2) Create Class Association for Dummy1 - explicit association for the class
    TsClassAssociation(_replace = True,
        py_canonical_name   = PyClass.name(Dummy1),
        ts_logical_name     = 'dummy1',
    ).save().throw()

    #-- 3) Create Class Association for Dummy2 - explicit association for the class
    TsClassAssociation(_replace = True,
        py_canonical_name   = PyClass.name(Dummy2),
        ts_logical_name     = 'dummy2',
    ).save().throw()

    #-- 4) Create an instance of Dummy1 and save it - must go to dummy1 TsStore
    d1 = Dummy1(text = 'hello dummy1')
    d1.save().throw()
    db1 = TsStore.instance_from_uri(extra_uris['dummy1'])
    with db1:
        assert Dummy1.exists_in_store(d1.id()), f'Dummy1 must have been associated with {extra_uris["dummy1"]}'

    #-- 5) Create an instance of Dummy3 and save it - must go to dummy2 TsStore because it deriuves from Dummy2 which is associated with dummy2 TsStore
    d3 = Dummy3(text = 'hello dummy3')
    d3.save().throw()
    db2 = TsStore.instance_from_uri(extra_uris['dummy2'])
    with db2:
        assert Dummy3.exists_in_store(d3.id()), f'Dummy3 must have been associated with {extra_uris["dummy2"]}'




from __future__ import annotations

from functools import total_ordering


@total_ordering
class ID:
    __slots__ = ('collection_name', 'value')

    def __init__(self, id_value: str = None, collection_name: str = None):
        self.value = id_value
        self.collection_name = collection_name

    def __eq__(self, other: ID):
        if not isinstance(other, ID):
            return NotImplemented
        return self.value == other.value and self.collection_name == other.collection_name

    def __bool__(self):
        return bool(self.value)

    def __repr__(self):
        return f'{self.value}' if not self.collection_name else f'{self.collection_name}/{self.value}'

    def __hash__(self):
        return hash((self.value, self.collection_name))

    def __lt__(self, other: ID) -> bool:
        if not isinstance(other, ID):
            return NotImplemented
        return (self.collection_name, self.value) < (other.collection_name, other.value)



import asyncio

import rio.testing
from ui_10x.rio.component_builder import DynamicComponent
from ui_10x.rio.widgets import CheckBox, GroupBox, Label, PushButton


async def test_group_box() -> None:
    """Test basic GroupBox widget functionality."""
    widget = GroupBox(title='Test Group')

    async with rio.testing.DummyClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # Test initial state
        assert widget['title'] == 'Test Group'

        # Test title change
        widget.set_title('Updated Group')
        await test_client.wait_for_refresh()
        assert widget['title'] == 'Updated Group'


async def test_group_box_with_children() -> None:
    """Test GroupBox widget with child widgets."""
    label = Label('Label inside group')
    checkbox = CheckBox('Checkbox inside group')
    button = PushButton('Button inside group')

    widget = GroupBox(None, 'Group with children', label, checkbox, button)

    async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # Test title
        assert widget['title'] == 'Group with children'

        # Test children
        children = widget.get_children()
        assert len(children) == 3
        assert label in children
        assert checkbox in children
        assert button in children
        assert 4 == await test_client.execute_js('document.querySelectorAll(".rio-column .rio-text").length')
        assert 1 == await test_client.execute_js('document.querySelectorAll(".rio-button").length')


async def test_group_box_add_remove_children() -> None:
    """Test GroupBox widget adding and removing children."""
    widget = GroupBox(title='Dynamic Group')
    label = Label('Dynamic Label')
    button = PushButton('Dynamic Button')

    async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # Test initial empty state
        assert len(widget.get_children()) == 0
        assert 1 == await test_client.execute_js('document.querySelectorAll(".rio-column .rio-text").length')
        assert 0 == await test_client.execute_js('document.querySelectorAll(".rio-button").length')

        # Test adding children
        widget.add_children(label)
        await test_client.wait_for_refresh()
        assert len(widget.get_children()) == 1
        assert label in widget.get_children()
        assert 2 == await test_client.execute_js('document.querySelectorAll(".rio-column .rio-text").length')
        assert 0 == await test_client.execute_js('document.querySelectorAll(".rio-button").length')

        widget.add_children(button)
        await test_client.wait_for_refresh()
        assert len(widget.get_children()) == 2
        assert button in widget.get_children()
        assert 3 == await test_client.execute_js('document.querySelectorAll(".rio-column .rio-text").length')
        assert 1 == await test_client.execute_js('document.querySelectorAll(".rio-button").length')

        # Test removing children
        widget._kwargs['children'] = [child for child in widget._kwargs['children'] if child is not label]
        widget.force_update()
        await test_client.wait_for_refresh()
        assert len(widget.get_children()) == 1
        assert label not in widget.get_children()
        assert button in widget.get_children()
        assert 2 == await test_client.execute_js('document.querySelectorAll(".rio-column .rio-text").length')
        assert 1 == await test_client.execute_js('document.querySelectorAll(".rio-button").length')

        widget._kwargs['children'] = [child for child in widget._kwargs['children'] if child is not button]
        widget.force_update()
        await test_client.wait_for_refresh()
        assert len(widget.get_children()) == 0
        assert 1 == await test_client.execute_js('document.querySelectorAll(".rio-column .rio-text").length')
        assert 0 == await test_client.execute_js('document.querySelectorAll(".rio-button").length')


# async def test_group_box_enabled_disabled() -> None:
#     """Test GroupBox enabled/disabled state."""
#     widget = GroupBox(title = 'Enabled Group')
#     label = Label('Child Label')
#     widget.add_widget(label)
#
#     async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
#         await asyncio.sleep(0.5)
#
#         # Test initial enabled state
#         assert widget['is_sensitive']
#         assert label.is_enabled()
#
#         # Test disabling group (should disable children)
#         widget.set_enabled(False)
#         await test_client.wait_for_refresh()
#         assert not widget['is_sensitive']
#         assert not label.is_enabled()
#
#         # Test re-enabling
#         widget.set_enabled(True)
#         await test_client.wait_for_refresh()
#         assert widget['is_sensitive']
#         assert label.is_enabled()


async def test_group_box_client_interaction() -> None:
    """Test GroupBox widget client interaction with timeout protection."""
    widget = GroupBox(title='Client Test')
    label = Label('Content')
    widget.add_children(label)

    async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # Test initial state
        assert widget['title'] == 'Client Test'
        assert len(widget.get_children()) == 1
        await check_title(widget, test_client)
        await check_label(label, test_client)

        # Test widget property changes
        widget.set_title('Updated Title')
        await test_client.wait_for_refresh()
        assert widget['title'] == 'Updated Title'

        # Verify client state reflects widget changes
        await check_title(widget, test_client)
        await check_label(label, test_client)

        # Test empty title
        widget.set_title('')
        await test_client.wait_for_refresh()
        assert widget['title'] == ''
        await check_title(widget, test_client)


async def check_label(label, test_client):
    assert 2 == await test_client.execute_js('document.querySelectorAll(".rio-column .rio-text").length')
    assert label['text'] == await test_client.execute_js('document.querySelectorAll(".rio-column .rio-text")[1].innerText')


async def check_title(widget, test_client):
    assert widget['title'] == await test_client.execute_js('document.querySelectorAll(".rio-column .rio-text")[0].innerText')



import typing

import rio


class Separator(rio.Component):
    """
    A component that draws a horizontal or vertical line to separate UI sections.

    Attributes:
        orientation: "horizontal" or "vertical" to define the separator's direction.
        thickness: The width of the line in Rio units (default: 0.1).
        color: The color of the line (default: gray).
    """

    orientation: typing.Literal['horizontal', 'vertical'] = 'horizontal'
    thickness: float = 0.1
    color: rio.Color = rio.Color.GRAY

    def build(self) -> rio.Component:
        if self.orientation == 'horizontal':
            return rio.Rectangle(min_height=self.thickness, fill=self.color, margin_y=0.5, grow_y=False, align_y=0.5)
        else:
            return rio.Rectangle(min_width=self.thickness, fill=self.color, margin_x=0.5, grow_x=False, align_x=0.5)



import atexit

from py10x_kernel import CORE_10X, PyLinkage, XCache


def cleanup():
    XCache.clear()
    PyLinkage.clear()

atexit.register(cleanup)

PyLinkage.init({
    CORE_10X.PACKAGE_NAME:                              __name__,

    CORE_10X.XNONE_MODULE_NAME:                         'xnone',
    CORE_10X.XNONE_CLASS_NAME:                          'XNone',

    CORE_10X.RC_MODULE_NAME:                            'rc',
    CORE_10X.RC_TRUE_NAME:                              'RC_TRUE',

    CORE_10X.NUCLEUS_MODULE_NAME:                       'nucleus',
    CORE_10X.NUCLEUS_CLASS_NAME:                        'Nucleus',

    CORE_10X.TRAITABLE_ID_MODULE_NAME:                  'traitable_id',
    CORE_10X.TRAITABLE_ID_CLASS_NAME:                   'ID',

    CORE_10X.TRAIT_METHOD_ERROR_MODULE_NAME:            'trait_method_error',
    CORE_10X.TRAIT_METHOD_ERROR_CLASS_NAME:             'TraitMethodError',

    CORE_10X.PACKAGE_REFACTORING_MODULE_NAME:           'package_refactoring',
    CORE_10X.PACKAGE_REFACTORING_CLASS_NAME:            'PackageRefactoring',
    CORE_10X.PACKAGE_REFACTORING_FIND_CLASS:            'find_class',
    CORE_10X.PACKAGE_REFACTORING_FIND_CLASS_ID:         'find_class_id',
})

try:
    from dev_10x.version import __version__
except ImportError:
    __version__ = '0.0.0+unknown'  # fallback for dev envs

from core_10x.environment_variables import EnvVars
if EnvVars.graph_on:
    from core_10x.exec_control import GRAPH_ON

    go = GRAPH_ON()
    go.begin_using()



import asyncio

import rio.testing.browser_client
from ui_10x.rio.component_builder import DynamicComponent
from ui_10x.rio.widgets import ListWidget


async def test_list_comprehensive() -> None:
    """Test ListWidget with comprehensive client-widget interaction verification."""
    widget = ListWidget()
    widget.add_items(['Item 1', 'Item 2', 'Item 3'])

    find_list_items = 'document.querySelectorAll(".rio-selectable-item")'
    find_selected_text = 'document.querySelector(".selected").querySelector(".rio-text").children[0].innerText'

    clicked_calls = []

    def clicked_handler(item):
        clicked_calls.append(item)

    widget.clicked_connect(clicked_handler)

    async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # 1) Verify client shows widget value (list items)
        assert widget.child_count() == 3
        client_items = await test_client.execute_js(f'{find_list_items}.length')
        assert client_items == 3

        # 2) Modify client value (user clicking first item)
        await test_client.click(10, 1)
        await asyncio.sleep(0.5)

        # 3) Verify widget reflects client value
        assert len(clicked_calls) == 1
        assert clicked_calls[0]['text'] == 'Item 1'
        assert widget['selected_items'] == [clicked_calls[0]['key']]

        # Verify client shows selection
        selected_text = await test_client.execute_js(find_selected_text)
        assert selected_text == 'Item 1'

        # Test clicking second item
        await test_client.click(10, test_client.window_height_in_pixels - 1)
        await asyncio.sleep(0.5)
        assert len(clicked_calls) == 2
        assert clicked_calls[1]['text'] == 'Item 3'  # Last item
        assert widget['selected_items'] == [clicked_calls[1]['key']]

        # Test widget changes propagate to client (clearing list)
        widget.clear()
        assert not widget.get_children()
        await test_client.wait_for_refresh()
        assert not widget.subcomponent.selected_items
        client_items = await test_client.execute_js(f'{find_list_items}.length')
        assert client_items == 0

        # Test adding item
        widget.add_item('New Item')
        await test_client.wait_for_refresh()
        assert widget.child_count() == 1
        client_items = await test_client.execute_js(f'{find_list_items}.length')
        assert client_items == 1

        # Test clicking new item
        await test_client.click(10, 1)
        await asyncio.sleep(0.5)
        assert len(clicked_calls) == 3
        assert clicked_calls[2]['text'] == 'New Item'


async def test_list_item_management() -> None:
    """Test ListWidget item management operations."""
    widget = ListWidget()
    widget.add_items(['A', 'B', 'C'])

    find_list_items = 'document.querySelectorAll(".rio-selectable-item")'

    async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # Test initial state
        assert widget.child_count() == 3
        client_items = await test_client.execute_js(f'{find_list_items}.length')
        assert client_items == 3

        # Test removing item
        removed_item = widget.take_item(1)  # Remove 'B'
        await test_client.wait_for_refresh()
        assert widget.child_count() == 2
        client_items = await test_client.execute_js(f'{find_list_items}.length')
        assert client_items == 2
        assert removed_item['text'] == 'B'

        # Test adding another item with timeout protection
        widget.add_item('Inserted')
        await test_client.wait_for_refresh()
        assert widget.child_count() == 3
        client_items = await test_client.execute_js(f'{find_list_items}.length')
        assert client_items == 3


async def test_list_disabled_interaction() -> None:
    """Test ListWidget disabled state blocks user interaction."""
    widget = ListWidget()
    widget.add_items(['Item 1', 'Item 2', 'Item 3'])

    clicked_calls = []

    def clicked_handler(item):
        clicked_calls.append(item)

    widget.clicked_connect(clicked_handler)

    async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # Test initial enabled state - clicks should work
        await test_client.click(10, 1)
        await asyncio.sleep(0.5)
        assert len(clicked_calls) == 1
        assert clicked_calls[0]['text'] == 'Item 1'
        assert widget['selected_items'] == [id(widget.get_children()[0])]

        # Disable widget
        widget.set_enabled(False)
        await test_client.wait_for_refresh()

        # Test that clicks are blocked when disabled
        initial_calls = len(clicked_calls)
        await test_client.click(10, 1)
        await asyncio.sleep(0.5)
        assert len(clicked_calls) == initial_calls  # No additional calls
        assert widget['selected_items'] == [id(widget.get_children()[0])]

        # Re-enable widget
        widget.set_enabled(True)
        await test_client.wait_for_refresh()

        # Test that clicks work again
        await test_client.click(10, 1)
        await asyncio.sleep(0.5)
        assert len(clicked_calls) == initial_calls + 1
        assert clicked_calls[-1]['text'] == 'Item 1'
        assert widget['selected_items'] == []



from __future__ import annotations

from contextlib import nullcontext
from dataclasses import dataclass
from typing import TYPE_CHECKING

from core_10x.concrete_traits import date_trait, dict_trait, flags_trait, list_trait
from core_10x.trait import T, Ui
from core_10x.traitable import traitable_trait

import ui_10x.concrete_trait_widgets  # do not remove -- this registers trait widgets
from ui_10x.choice import MultiChoice
from ui_10x.py_data_browser import PyDataBrowser
from ui_10x.trait_widget import TraitWidget
from ui_10x.utils import UxDialog, UxStyleSheet, ux, ux_pick_date

if TYPE_CHECKING:
    from collections.abc import Callable

    from core_10x.traitable import Traitable
    from core_10x.exec_control import BTP
    from core_10x.rc import RC
    from core_10x.trait import Trait


@dataclass
class TraitableWrapper:
    traitable: Traitable
    traitable_processor: Callable[[], BTP]

    def set_value(self, trait: Trait, value) -> RC:
        with self.traitable_processor() or nullcontext():
            return self.traitable.set_value(trait, value)

    def get_value(self, trait: Trait):
        with self.traitable_processor() or nullcontext():
            value = self.traitable.get_value(trait)
            return value

    def is_valid(self, trait: Trait) -> bool:
        with self.traitable_processor() or nullcontext():
            return self.traitable.is_valid(trait)

    def invalidate_value(self, trait: Trait) -> None:
        with self.traitable_processor() or nullcontext():
            return self.traitable.invalidate_value(trait)

    def get_style_sheet(self, trait: Trait):
        with self.traitable_processor() or nullcontext():
            return self.traitable.get_style_sheet(trait)

    def get_choices(self, trait: Trait):
        with self.traitable_processor() or nullcontext():
            return self.traitable.get_choices(trait)

    def create_ui_node(self, trait: Trait, callback):
        with self.traitable_processor() or nullcontext():
            return self.traitable.bui_class().create_ui_node(self.traitable, trait, callback)

    def update_ui_node(self, trait: Trait):
        with self.traitable_processor() or nullcontext():
            return self.traitable.bui_class().update_ui_node(self.traitable, trait)


class TraitEditor:
    def __init__(self, traitable, trait: Trait, ui_hint: Ui, custom_callback: Callable[[], None] = None, traitable_processor: Callable[[], BTP] = None):
        self.traitable = TraitableWrapper(traitable, traitable_processor)
        self.trait = trait
        self.widget: TraitWidget | None = None
        self.ui_hint = ui_hint
        self.trait_callback = self._establish_callback(custom_callback)

    def is_read_only(self) -> bool:
        return self.ui_hint.flags_on(Ui.READ_ONLY)

    def _establish_callback(self, custom_callback: Callable[[], None]):
        if custom_callback:
            return custom_callback

        if self.is_read_only():
            return None

        return self.builtin_callback()

    def new_label(self) -> ux.Widget:
        if self.ui_hint.param('right_label', False):
            return ux.Label(' ')

        if self.trait_callback:
            label = ux.PushButton()
            label.set_flat(True)
            label.set_style_sheet('text-align: left')
            label.set_text(self.ui_hint.label + '...')
            label.clicked_connect(lambda v: self.trait_callback(self))
            if self.trait.flags_on(T.EXPENSIVE):
                UxStyleSheet.modify(label, {Ui.FG_COLOR: 'purple', Ui.FONT_STYLE: 'italic'})
                label.set_tool_tip('Reaction on clicking this button may take considerable time...')
        else:
            label = ux.Label(self.ui_hint.label)

        return label

    def new_widget(self, update_self=True) -> TraitWidget:
        tw: TraitWidget = TraitWidget.instance(self)
        assert tw, f'{self.traitable.traitable.__class__}.{self.trait.name} - unknown trait widget class'

        avg_char_width = tw.font_metrics().average_char_width()
        min_width = self.ui_hint.param('min_width', 0)
        if min_width > 0:
            tw.set_minimum_width(min_width * avg_char_width)

        max_width = self.ui_hint.param('max_width', 0)
        if max_width > 0:
            tw.set_maximum_width(max_width * avg_char_width)

        if update_self:
            self.widget = tw

        return tw

    # ---- Built-in Callbacks

    def date_cb(self):
        ux_pick_date(
            title=f'Pick a date for {self.ui_hint.label}',
            show_date=self.traitable.get_value(self.trait),
            on_accept=lambda value: self.traitable.set_value(self.trait, value),
        )

    def list_cb(self):
        choices = self.traitable.get_value(self.trait)
        mc = MultiChoice(choices=choices)
        w = mc.widget()
        if not w:
            return

        UxDialog(
            w,
            title=f'Choose one or more values for {self.ui_hint.label}',
            accept_callback=lambda ctx: self.traitable.set_value(self.trait, mc.values_selected),
        ).show()

    def dict_cb(self):
        data = dict(self.traitable.get_value(self.trait))
        rc = PyDataBrowser.edit(data, title=f'Edit {self.trait.ui_hint.label}')  # TODO: callback
        if rc:
            self.traitable.set_value(self.trait, data)

    def flags_cb(self):
        flags = self.traitable.get_value(self.trait)
        bits_class = self.trait.data_type
        tags_selected = bits_class.data_type.names_from_value(flags)
        mc = MultiChoice(*tags_selected, choices=bits_class.choose_from())
        w = mc.widget()
        if not w:
            return

        UxDialog(
            w,
            title=f'Choose one or more flags for {self.ui_hint.label}',
            accept_callback=lambda ctx: self.traitable.set_value(self.trait, mc.values_selected),
        ).show()

    def traitable_cb(self):
        # -- EntityEditor - popup
        ...

    def expensive_cb(self):
        value = self.traitable.get_value(self.trait)
        self.widget.set_widget_value(value)
        self.widget.style_sheet.restore()

    def builtin_callback(self):
        cb = self.s_builtin_callbacks.get(self.trait.__class__)
        if not cb:
            if self.trait.flags_on(T.EXPENSIVE):
                cb = self.__class__.expensive_cb

        return cb

    # fmt: off
    s_builtin_callbacks = {
        date_trait:         date_cb,
        list_trait:         list_cb,
        dict_trait:         dict_cb,
        flags_trait:        flags_cb,
        traitable_trait:    traitable_cb,
    }
    # fmt: on



from core_10x.traitable import Traitable, Trait, T, RT, RC, XNone

class TraitableHeir(Traitable):
    _grantor: Traitable = T()

    s_grantor_traits = {}
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        cls.s_grantor_traits = dir = {}
        grantor_trait = cls.trait('_grantor')
        trait: Trait
        for trait in cls.traits():
            if trait is grantor_trait or trait.flags_on(T.RESERVED):
                continue
            if not trait.has_custom_getter() and trait.default is XNone:   #-- trait has neither getter nor default value
                trait.set_f_get(lambda self, trait_name = trait.name: self.heir_getter(trait_name), False)
                dir[trait.name] = trait

    def heir_getter(self, trait_name: str):
        grantor = self._grantor
        if not grantor:
            return XNone
        trait = grantor.__class__.trait(trait_name)
        return grantor.get_value(trait) if trait else XNone

    def serialize_object(self, save_references = False) -> dict:
        serialized_data = super().serialize_object(save_references = save_references)
        for name, trait in self.__class__.s_grantor_traits.items():
            if not self.is_set(trait):
                del serialized_data[name]

        return serialized_data



from __future__ import annotations

import importlib
from typing import TYPE_CHECKING, Literal

from ibis.common.exceptions import TableNotFound

from core_10x.resource import REL_DB, Resource, ResourceSpec

if TYPE_CHECKING:
    import ibis.expr.types as ir
    import polars as pl


class RelDb(Resource, resource_type=REL_DB):

    def __init__(self, uri: str):
        self._uri = uri
        self._connection = None

    @classmethod
    def instance(cls, *args, **kwargs) -> RelDb:
        if args:
            raise ValueError('RelDb.instance accepts keyword arguments only.')
        if dbname := kwargs.get(cls.DBNAME_TAG):
            kwargs[cls.DBNAME_TAG] = dbname.replace('\\', '/')  # normalize Windows paths; backslashes are illegal in URIs
        return cls(ResourceSpec(cls, kwargs).uri())

    def __enter__(self):
        self.begin_using()
        return self

    def on_enter(self):
        try:
            ibis = importlib.import_module('ibis')
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError('ibis is required for RelDb connections.') from e
        self._connection = ibis.connect(self._uri)

    def on_exit(self):
        if self._connection is not None:
            self._connection.disconnect()
            self._connection = None

    def query(self, table_name: str) -> ir.Table|None:
        """Return an ibis Table for the given table, for further ibis manipulation."""
        assert self._connection is not None, 'RelDb must be used as a context manager.'
        try:
            return self._connection.table(table_name)
        except TableNotFound:
            return None

    def insert(self, table_name: str, df: pl.DataFrame, if_exists: Literal['replace', 'append', 'fail'] = 'fail') -> None:
        """Insert a polars DataFrame into a table via Polars ADBC.

        if_exists: 'fail' (default) raises if the table exists;
                   'replace' drops and recreates; 'append' adds rows.
        """
        assert self._connection is not None, 'RelDb must be used as a context manager.'
        df.write_database(table_name, self._uri, engine='adbc', if_table_exists=if_exists)

    def drop_table(self, table_name: str) -> None:
        assert self._connection is not None, 'RelDb must be used as a context manager.'
        self._connection.drop_table(table_name)



if __name__ == '__main__':
    from core_10x.code_samples.person import Person

    to_save = True


    #print(db.collection_names())

    #Person.collection().delete('Sasha|Davidovich')
    if to_save:
        p = Person(first_name = 'Sasha', last_name = 'Davidovich')
        p.weight_lbs=200
        rev = p.save()

    else:
        p = Person(first_name = 'Sasha', last_name = 'Davidovich')
        p.weight_lbs=200

        print(p.age)

        p.age = 61
        rev = p.save()
        print(p.age)

        p.age = -10
        print(p.age)

        p.reload()
        print(p.age)









import pytest

from core_10x.environment_variables import EnvVars


@pytest.fixture(params=[True, False], ids=['with_transactions', 'without_transactions'])
def with_transactions(request, ts_instance, monkeypatch):
    use_transactions = request.param
    if use_transactions and not ts_instance.supports_transactions():
        pytest.skip('Store does not support transactions')

    monkeypatch.setenv('XX_USE_TS_STORE_TRANSACTIONS', '1' if use_transactions else '0')
    object.__getattribute__(EnvVars, 'use_ts_store_transactions').fget.clear()
    yield use_transactions



from core_10x.traitable import VaultUser, VaultResourceAccessor, TsStore
from core_10x.vault_utils import VaultUtils
from core_10x.environment_variables import EnvVars

from infra_10x.mongodb_store import MongoStore

if __name__ == '__main__':
    #EnvVars.vault_ts_store_uri

    #vault = MongoStore.instance(hostname = 'localhost', dbname = 'vault', port = 27018, username = 'AMD', password = '')
    VaultUtils.new_vault()



from __future__ import annotations

from infra_10x.mongodb_store import MongoStore  # TODO: backbone
from ui_10x.rio.component_builder import UserSessionContext

import rio


@rio.page(
    name='Login/Logout',
    url_segment='',
)
class LoginPage(rio.Component):
    username: str = ''
    password: str = ''

    error_message: str = ''

    _currently_logging_in: bool = False

    async def login(self, _: rio.TextInputConfirmEvent | None = None) -> None:
        try:
            self._currently_logging_in = True
            self.force_refresh()

            try:
                runtime_context = self.session[UserSessionContext]
                running, with_auth = MongoStore.is_running_with_auth(runtime_context.host)
                if not running:
                    self.error_message = 'Authentication is not available'
                    return
                if not self.username and with_auth:
                    self.error_message = 'Username is required'
                    return
                runtime_context.traitable_store = MongoStore.instance(
                    hostname=runtime_context.host, dbname=runtime_context.dbname, username=self.username, password=self.password
                )
                runtime_context.authenticated = True

            except Exception as e:
                self.error_message = f'Login error - try again\n{e}'
                return

            # The login was successful
            self.error_message = ''

        # Done
        finally:
            self._currently_logging_in = False

    async def logout(self, _: rio.TextInputConfirmEvent | None = None) -> None:
        try:
            runtime_context = self.session[UserSessionContext]
            runtime_context.authenticated = False
            runtime_context.mongo_store = None
        except Exception as e:
            self.error_message = f'Logout error\n {e}'
            return

        # The login was successful
        self.error_message = ''

    def build(self) -> rio.Component:
        rows = (
            [
                rio.TextInput(text=self.bind().username, label='Username', on_confirm=self.login),
                rio.TextInput(text=self.bind().password, label='Password', is_secret=True, on_confirm=self.login),
                rio.Button('Sign In', on_press=self.login, is_loading=self._currently_logging_in),
            ]
            if not self.session[UserSessionContext].authenticated
            else [
                rio.Button('Sign Out', on_press=self.logout),
            ]
        )

        return rio.Card(
            rio.Column(
                rio.Text('Sign In/Sign Out', style='heading1', justify='center'),
                rio.Banner(text=self.error_message, style='danger', margin_top=1),
                *rows,
                spacing=1,
                margin=2,
            ),
            margin_x=0.5,
            align_y=0.5,
            align_x=0.5,
            min_width=24,
        )



from __future__ import annotations

import typing as t

import rio


class Splitter(rio.Component):
    """
    A custom Rio component that arranges children horizontally like a Row,
    or vertically as a column with draggable splitters between them for resizing.
    """

    # Props
    children: list[rio.Component] = []
    direction: t.Literal['horizontal', 'vertical'] = 'vertical'
    handle_size: float = 0.25  # Width of the splitter handle
    min_size_percent: float = 10.0  # Minimum width for each child (%)
    child_proportions: t.Literal['homogeneous'] | t.Sequence[float] = 'homogeneous'
    _component_width: float = 0.0
    _component_height: float = 0.0

    def __post_init__(self):
        if not isinstance(self.child_proportions, (list, tuple)):
            num_children = len(self.children)
            self.child_proportions = [1.0] * num_children if num_children else []
        else:
            assert len(self.child_proportions) == len(self.children)
            assert all(p >= 0 for p in self.child_proportions), 'Proportions must be non-negative'
            self.child_proportions = list(self.child_proportions)

    def on_drag(self, index: int, event: rio.PointerMoveEvent) -> None:
        """
        Handle drag events on the splitter at the given index.
        Adjusts the proportions of the two adjacent children.
        """
        horizontal = self.direction == 'horizontal'
        total_size = self._component_width if horizontal else self._component_height

        # Convert drag movement to a proportion change
        total_proportion = sum(self.child_proportions)
        relative_size = event.relative_x if horizontal else event.relative_y
        delta_proportion = (relative_size / total_size) * total_proportion

        # Adjust the proportions of the left and right children
        prev_index = index
        next_index = index + 1

        # Calculate current sizes as percentages to check minimum constraints
        current_sizes = [(p / total_proportion * 100.0) for p in self.child_proportions]
        new_left_size = current_sizes[prev_index] + (delta_proportion / total_proportion * 100.0)
        new_right_size = current_sizes[next_index] - (delta_proportion / total_proportion * 100.0)

        # Check minimum size constraints
        if new_left_size >= self.min_size_percent and new_right_size >= self.min_size_percent:
            self.child_proportions[prev_index] += delta_proportion
            self.child_proportions[next_index] -= delta_proportion
            # Ensure proportions don't go negative
            self.child_proportions[prev_index] = max(0.0, self.child_proportions[prev_index])
            self.child_proportions[next_index] = max(0.0, self.child_proportions[next_index])

        self.child_proportions = self.child_proportions  # force refresh

    def build(self) -> rio.Component:
        # If no children, return an empty component
        if not self.children:
            return rio.Rectangle()

        # Build the layout with children and splitters
        components = []
        horizontal = self.direction == 'horizontal'
        for i, child in enumerate(self.children):
            # Wrap child in ScrollArea
            scrollable_content = rio.ScrollContainer(
                content=child,
                # scroll_x='never' if horizontal else 'auto',
                # scroll_y='auto' if horizontal else 'never',
            )
            # Create the pane
            pane = rio.Rectangle(
                content=scrollable_content,
                **{'grow_x' if horizontal else 'grow_y': True},  # Stretch to fill proportional space
                margin=1,  # Spacing around the child content
            )
            # Add a splitter handle to the right of all but the last pane
            if i < len(self.children) - 1:
                splitter = rio.PointerEventListener(
                    content=rio.Rectangle(
                        **{'grow_x' if horizontal else 'grow_y': False, 'min_width' if horizontal else 'min_height': self.handle_size},
                        fill=rio.Color.from_hex('#808080'),
                        cursor='move',  # Valid CursorStyle for dragging
                    ),
                    on_drag_move=lambda event, idx=i: self.on_drag(idx, event),
                    **{
                        'align_x' if horizontal else 'align_y': 1.0,  # Position at the right edge
                        'margin_right' if horizontal else 'margin_bottom': -self.handle_size / 2,  # Extend slightly into the next pane
                    },
                )
                # Combine pane and splitter in a Stack
                components.append(
                    rio.Stack(
                        pane,
                        splitter,
                        **{'grow_x' if horizontal else 'grow_y': True},  # Ensure the Stack follows the proportion
                    )
                )
            else:
                # Last pane has no splitter
                components.append(pane)

        container = rio.Row if self.direction == 'horizontal' else rio.Column
        return container(
            *components,
            spacing=0,
            proportions=self.bind().child_proportions,  # Dynamically control pane sizes
        )

    @rio.event.on_resize
    def _on_resize(self, event: rio.event.ComponentResizeEvent) -> None:
        self._component_width = event.width
        self._component_height = event.height



from core_10x.code_samples.person import Person
from core_10x.ts_store import TsStore

from ui_10x.traitable_editor import TraitableEditor, TraitableView

if __name__ == '__main__':
    db = TsStore.instance_from_uri('mongodb://localhost/test')
    db.begin_using()

    p = Person(first_name = 'Sasha', last_name = 'Davidovich')
    view = TraitableView.modify(Person,
        #weight_qu   = UiMod(flags = Ui.HIDDEN),
    )
    e = TraitableEditor.editor(p, view = view)
    rc = e.dialog(save=True).exec()






"""Store-agnostic TsStore transaction tests. Use ts_instance fixture from conftest (core_10x → TestStore, infra_10x → MongoStore)."""
from uuid import uuid4

import pytest

from core_10x.rc import RC
from core_10x.testlib.fixtures import with_transactions
from core_10x.trait_definition import T
from core_10x.traitable import Traitable
from core_10x.ts_store import SaveIfChanged


class TestTsStoreTransaction:
    """Transaction semantics: commit applies changes, abort discards. Runs against ts_instance from conftest."""

    @pytest.fixture(autouse=True)
    def _skip_if_store_does_not_support_transactions(self, ts_instance):
        """Skip this test class when the store does not support transactions (e.g. MongoDB standalone)."""
        if not getattr(ts_instance, 'supports_transactions', lambda: True)():
            pytest.skip(
                "Store does not support transactions "
                "(e.g. MongoDB standalone; needs replica set or mongos)"
            )

    @pytest.fixture
    def store(self, ts_instance):
        return ts_instance

    @pytest.fixture
    def coll_name(self):
        return f'ts_store_tx_test#{uuid4().hex}'

    @pytest.fixture
    def coll(self, store, coll_name):
        c = store.collection(coll_name)
        yield c
        if coll_name in store.collection_names():
            store.delete_collection(coll_name)

    def test_transaction_yields_tx_with_commit_abort(self, store):
        with store.transaction() as tx:
            assert hasattr(tx, 'commit') and hasattr(tx, 'abort')
            assert not tx.ended

    def test_commit_applies_pending_writes(self, store, coll):
        coll.save_new({'_id': 'a', '_rev': 0}, overwrite=False)
        assert coll.count() == 1
        with store.transaction() as tx:
            coll.save_new({'_id': 'b', '_rev': 0}, overwrite=False)
            assert coll.count() == 2
            tx.commit()
        assert coll.count() == 2
        assert coll.id_exists('b')

    def test_abort_discards_pending_writes(self, store, coll):
        coll.save_new({'_id': 'a', '_rev': 0}, overwrite=False)
        with store.transaction() as tx:
            coll.save_new({'_id': 'b', '_rev': 0}, overwrite=False)
            tx.abort()
        assert coll.count() == 1
        assert not coll.id_exists('b')

    def test_exception_triggers_abort(self, store, coll):
        coll.save_new({'_id': 'a', '_rev': 0}, overwrite=False)
        with pytest.raises(RuntimeError, match='rollback'):
            with store.transaction():
                coll.save_new({'_id': 'd', '_rev': 0}, overwrite=False)
                raise RuntimeError('rollback')
        assert coll.count() == 1
        assert not coll.id_exists('d')

    def test_manual_commit_before_exit(self, store, coll):
        with store.transaction() as tx:
            coll.save_new({'_id': 'x', '_rev': 0}, overwrite=False)
            tx.commit()
        assert coll.count() == 1
        assert coll.id_exists('x')

    def test_double_commit_no_op(self, store, coll):
        with store.transaction() as tx:
            coll.save_new({'_id': 'y', '_rev': 0}, overwrite=False)
            tx.commit()
            tx.commit()
        assert coll.count() == 1

    def test_transaction_sees_own_pending_in_find(self, store, coll):
        with store.transaction() as tx:
            coll.save_new({'_id': 'z', '_rev': 0}, overwrite=False)
            docs = list(coll.find())
            assert len(docs) == 1
            assert docs[0]['_id'] == 'z'
            tx.commit()
        assert coll.count() == 1

    def test_transaction_delete_pending(self, store, coll):
        coll.save_new({'_id': 'del', '_rev': 0}, overwrite=False)
        with store.transaction() as tx:
            assert coll.delete('del')
            assert not coll.id_exists('del')
            tx.commit()
        assert coll.count() == 0
        assert not coll.id_exists('del')

    def test_transaction_delete_then_abort_restores_visibility(self, store, coll):
        coll.save_new({'_id': 'del2', '_rev': 0}, overwrite=False)
        with store.transaction() as tx:
            coll.delete('del2')
            assert not coll.id_exists('del2')
            tx.abort()
        assert coll.id_exists('del2')
        assert coll.count() == 1


class TestSaveIfChanged:
    @pytest.fixture
    def coll_names(self,ts_instance):
        coll_names = tuple(f'save_if_changed#{x}#{uuid4().hex}' for x in ('a','b'))
        assert not set(coll_names).intersection(ts_instance.collection_names())
        yield coll_names
        for coll_name in coll_names:
            ts_instance.delete_collection(coll_name)

    def test_save_if_changed_filters_by_classes(self, ts_instance, coll_names, with_transactions):  # noqa: F811

        coll_a_name, coll_b_name = coll_names
        class TrackedA(Traitable, custom_collection=True):
            i: int = T(T.ID)
            value: int = T()

        class TrackedB(Traitable, custom_collection=True):
            i: int = T(T.ID)
            value: int = T()

            def save(self,save_references=True):
                return RC(False,'boom')

        with ts_instance:
            a = TrackedA(i=1, _collection_name=coll_a_name)
            b = TrackedB(i=2, _collection_name=coll_b_name)
            with SaveIfChanged([TrackedA]) as tracker:
                a.value = 10
                b.value = 20
                assert tracker.tracked_objects() == [a,b]

            assert ts_instance.collection(coll_a_name).count() == 1
            assert ts_instance.collection(coll_b_name).count() == 0

            a.delete()
            assert ts_instance.collection(coll_a_name).count() == 0

            with pytest.raises(RuntimeError,match='boom'):
                with SaveIfChanged() as tracker:
                    a.value = 20
                    b.value = 30
                assert tracker.tracked_objects() == [a,b]

            assert ts_instance.collection(coll_a_name).count() == int(not with_transactions)
            assert ts_instance.collection(coll_b_name).count() == 0

    def test_save_if_changed_requires_storable_classes(self):
        class NotStorable:
            @classmethod
            def is_storable(cls):
                return False

        with pytest.raises(RuntimeError, match='SaveIfChanged must be storable'):
            with SaveIfChanged([NotStorable]):
                pass



from __future__ import annotations

import os
import shutil
import subprocess
import sys
from pathlib import Path
import importlib.metadata as md

from tomlkit import document, dumps, inline_table, table, TOMLDocument

ROOTS = {
    'py10x-core': {'path': '../py10x', 'editable': True},
    'py10x-kernel': {'path': '../cxx10x/core_10x', 'editable': True},
    'py10x-infra': {'path': '../cxx10x/infra_10x', 'editable': True},
}

REPOS = {
    'py10x-core': {
        'git': 'https://github.com/10x-software/py10x.git',
        'branch': 'main',
    },
    'py10x-kernel': {
        'git': 'https://github.com/10x-software/cxx10x.git',
        'branch': 'main',
        'subdirectory': 'core_10x',
    },
    'py10x-infra': {
        'git': 'https://github.com/10x-software/cxx10x.git',
        'branch': 'main',
        'subdirectory': 'infra_10x',
    },
}

PROFILES = {
    'user': {},
    'domain-dev': REPOS,
    'py10x-dev': REPOS | {'py10x-core': ROOTS['py10x-core']},
    'py10x-core-dev': ROOTS,
}


def source_version(src: str) -> str:
    out = subprocess.check_output(
        [sys.executable, '-m', 'setuptools_scm'],
        cwd=src,
        text=True,
    )
    return out.strip()


def should_reinstall(dist_name: str, src: str, verbose=False, quiet=False) -> bool:
    if 'cxx10x' not in src:
        if verbose:
            print(f'No need to reinstall {dist_name} - editable installs work for python packages')
        return False

    try:
        installed = md.version(dist_name)
        new_version = source_version(src)
    except md.PackageNotFoundError as e:
        if not quiet:
            print(f'Will reinstall {dist_name} just in case: got error {e}')
        return True

    will_reinstall = installed != new_version
    if (will_reinstall or verbose) and not quiet:
        print(f'{"Will reinstall" if will_reinstall else "No need to reinstall"} {dist_name}: old_version={installed} new_version={new_version}')
    return will_reinstall


def get_extra_options(profile: str):
    return [
        f'--reinstall-package={package}'
        for package, source in PROFILES[profile].items()
        if source is ROOTS.get(package) and should_reinstall(package, source['path'])
    ]


def uv_sources_block(user_profile: str) -> TOMLDocument | None:
    uv_sources = PROFILES[user_profile]
    if not uv_sources:
        return None

    doc = document()
    doc['tool'] = tool_tbl = table()
    tool_tbl['uv'] = uv_tbl = table()
    uv_tbl['sources'] = sources_tbl = table()

    for package, source in uv_sources.items():
        it = inline_table()
        # Stable key order for readability
        for k in sorted(source.keys()):
            it[k] = source[k]
        sources_tbl[package] = it

    return doc


def uv_sync(user_profile: str, *args):
    project_root = Path('.').resolve()
    pyproject = project_root / 'pyproject.toml'

    if not pyproject.exists():
        raise RuntimeError('pyproject.toml not found')

    py_bak = pyproject.with_suffix('.toml.bak')

    if py_bak.exists():
        raise RuntimeError(f'Backup already exists: {py_bak}')

    src_block = uv_sources_block(user_profile)
    try:
        if src_block is not None:
            shutil.copy2(pyproject, py_bak)
            with pyproject.open('a', encoding='utf-8', newline='\n') as f:
                f.write('\n' + src_block.as_string())
        subprocess.run(['uv', 'sync', *args], cwd=project_root, check=True)

    finally:
        if src_block is not None and py_bak.exists():
            shutil.copy2(py_bak, pyproject)
            py_bak.unlink(missing_ok=True)


def ensure_chromium_installed() -> None:
    try:
        import playwright  # check package exists
    except ImportError:
        return  # playwright not installed → skip

    from playwright.sync_api import sync_playwright, Error as PlaywrightError

    try:
        with sync_playwright() as p:
            p.chromium.launch(headless=True)  # probe
        return  # already good
    except PlaywrightError:
        print('Installing Playwright Chromium...')
        subprocess.run(['playwright', 'install', 'chromium'], check=True)
        print('Done.')


def main():
    if len(sys.argv) < 2 or sys.argv[1] not in PROFILES:
        print(f'Usage: uvx --from "10x-core[dev]" uv_sync {"|".join(PROFILES)} [uv sync options (see uv sync --help for details)]')
    else:
        profile = sys.argv[1]
        sources = uv_sources_block(profile)
        s = dumps(sources['tool']['uv']['sources']) if sources else ''
        print(f'Using the {"following" if s else "default"} sources for `{profile}` profile{":" if s else "."}\n{s}')
        if opts := get_extra_options(profile):
            print(f'Using the following extra options for `{profile}` profile: {opts}')

        uv_sync(profile, *sys.argv[2:], *opts)
        if profile == 'py10x-core-dev':
            ensure_chromium_installed()


if __name__ == '__main__':
    main()



from core_10x.xnone import XNone


class TraitMethodError(Exception):
    """
    NOTE: other_exc must be set in except clause ONLY!
    """

    @staticmethod
    def create(traitable, traitable_class, trait_name: str, method_name: str, value=XNone, other_exc: Exception = None, args=()):
        assert other_exc

        if isinstance(other_exc, TraitMethodError):
            return other_exc

        msg = []
        traitable_class = traitable_class or traitable.__class__
        msg.append(f'Failed in {traitable_class}.{trait_name}.{method_name}')
        if traitable:
            msg.append(f'    object = {traitable.id()};')

        if value is not XNone:
            msg.append(f'    value = {value}')

        if args:
            msg.append(f'    args = {args}')

        if other_exc:
            msg.append(f'original exception = {type(other_exc).__name__}: {other_exc!s}')

        exc = TraitMethodError('\n'.join(msg))

        if other_exc and (tb := other_exc.__traceback__):
            exc = exc.with_traceback(tb)

        return exc



import copy
import functools

from py10x_kernel import BTraitFlags as T

from core_10x.named_constant import Enum


class UiHint:
    # fmt: off
    #-- Flags
    HIDDEN      = 0x1
    READ_ONLY   = 0x2
    SELECT_ONLY = 0x4
    SEPARATOR   = 0x8

    #-- Stylesheet Attributes
    FG_COLOR        = 'color'
    BG_COLOR        = 'background-color'
    FONT            = 'font-family'
    FONT_STYLE      = 'font-style'
    FONT_WEIGHT     = 'font-weight'
    BORDER_WEIGHT   = 'border-weight'
    BORDER_WIDTH    = 'border-width'
    BORDER_COLOR    = 'border-color'
    BORDER_STYLE    = 'border-style'

    class WIDGET_TYPE(Enum):
        NONE        = ()
        LINE        = ()
        TEXT        = ()
        TEXT4LIST   = ()
        PASSWORD    = ()
        CHECK       = ()
        CHOICE      = ()
        PIXMAP      = ()
        PUSH        = ()
        FILE        = ()
    # fmt: on
    @staticmethod
    def partial(widget_type_fn, **params):
        return functools.partial(widget_type_fn, **params)

    def __init__(self, label: str = None, flags: int = 0x0, tip: str = None, widget_type: WIDGET_TYPE = None, **params):
        self.label: str = label
        self.flags: int = flags
        self.tip: str = tip
        self.widget_type = widget_type
        self.params: dict = params

    def adjust(self, trait):
        if self.label is None:
            self.label = ' '.join(p.capitalize() for p in trait.name.split('_'))

        if self.tip is None:
            self.tip = self.label

        if self.widget_type is None:
            self.widget_type = trait.s_ui_hint.widget_type

        trait_flags = trait.flags
        if trait_flags & T.HIDDEN.value():
            self.flags |= self.HIDDEN
        if trait_flags & T.READONLY.value():
            self.flags |= self.READ_ONLY

    def flags_on(self, flags: int) -> bool:
        return bool(self.flags & flags)

    def set_reset_flags(self, to_set: int, to_reset: int = 0x0):
        self.flags = (self.flags | to_set) & ~to_reset

    def param(self, param_name: str, default_value):
        return self.params.get(param_name, default_value)

    @staticmethod
    def line(label: str = None, flags: int = 0x0, tip: str = None, min_width: int = 1, align_h: int = 1, **kwargs):
        return Ui(label=label, flags=flags, tip=tip, widget_type=Ui.WIDGET_TYPE.LINE, min_width=min_width, align_h=align_h, **kwargs)

    @staticmethod
    def text(label: str = None, flags: int = 0x0, tip: str = None, min_width: int = 1, **kwargs):
        return Ui(label=label, flags=flags, tip=tip, widget_type=Ui.WIDGET_TYPE.TEXT, min_width=min_width, **kwargs)

    @staticmethod
    def text4list(label: str = None, flags: int = 0x0, tip: str = None, min_width: int = 1, **kwargs):
        return Ui(label=label, flags=flags, tip=tip, widget_type=Ui.WIDGET_TYPE.TEXT4LIST, min_width=min_width, **kwargs)

    @staticmethod
    def password(label: str = None, flags: int = 0x0, tip: str = None, min_width: int = 10, **kwargs):
        return Ui(label=label, flags=flags, tip=tip, widget_type=Ui.WIDGET_TYPE.PASSWORD, min_width=min_width, **kwargs)

    @staticmethod
    def check(label: str = None, flags: int = 0x0, tip: str = None, right_label: bool = False, align_h: int = 1, **kwargs):
        return Ui(label=label, flags=flags, tip=tip, widget_type=Ui.WIDGET_TYPE.CHECK, right_label=right_label, align_h=align_h, **kwargs)

    @staticmethod
    def choice(label: str = None, flags: int = 0x0, tip: str = None, align_h: int = 1, **kwargs):
        return Ui(label=label, flags=flags, tip=tip, widget_type=Ui.WIDGET_TYPE.CHOICE, align_h=align_h, **kwargs)

    @staticmethod
    def pixmap(label: str = None, flags: int = 0x0, tip: str = None, w: int = 10, h: int = 10, **kwargs):
        return Ui(label=label, flags=flags, tip=tip, widget_type=Ui.WIDGET_TYPE.PIXMAP, w=w, h=h, **kwargs)

    @staticmethod
    def button(label: str = None, flags: int = 0x0, tip: str = None, min_width: int = 1, **kwargs):
        return Ui(label=label, flags=flags, tip=tip, widget_type=Ui.WIDGET_TYPE.PUSH, min_width=min_width, **kwargs)

    @staticmethod
    def file(label: str = None, flags: int = 0x0, tip: str = None, min_width: int = 1, **kwargs):
        return Ui(label=label, flags=flags, tip=tip, widget_type=Ui.WIDGET_TYPE.FILE, min_width=min_width, **kwargs)


class UiHintModification(UiHint):
    # fmt: off
    def __init__(
        self,
        label: str                      = None,
        flags: int                      = None,
        tip: str                        = None,
        widget_type: UiHint.WIDGET_TYPE = None,
        **params,
    ):
        # fmt: on
        super().__init__(label, flags = flags, tip = tip, widget_type = widget_type, **params)

    def apply(self, ui_hint: UiHint) -> UiHint:
        hint = copy.deepcopy(ui_hint)
        if self.label is not None:
            hint.label = self.label

        flags = self.flags
        if flags is not None:
            if isinstance(flags, int):      #-- to_set
                hint.flags |= flags
            elif isinstance(flags, tuple):  #-- (to_set, to_reset)
                hint.flags = (hint.flags | flags[0]) & ~flags[1]

        if self.tip is not None:
            hint.tip = self.tip

        if self.widget_type is not None:
            hint.widget_type = self.widget_type

        hint.params.update(self.params)

        return hint


# fmt: off
UiHint.NONE = UiHint(widget_type = UiHint.WIDGET_TYPE.NONE)
Ui          = UiHint
UiMod       = UiHintModification
# fmt: on





if __name__ == '__main__':
    #import os;assert os.environ.setdefault('UI_PLATFORM', 'Rio') == os.getenv('UI_PLATFORM')

    from ui_10x.utils import ux, ux_answer, ux_success, ux_warning

    app = ux.init()

    ret = ux_answer('This is a message box. '*50+'Ok?', title='Message Box Title')

    print(ret)


    ret = ux_success('This is a success message box. '*50+'Ok?', title='Success Message Box Title')
    print(ret)

    ret = ux_warning('This is a warning message box. '*50+'Ok?', title='Warning Message Box Title')
    print(ret)




import py10x_infra
import pytest

import infra_10x


@pytest.mark.parametrize('pattern', ['0.0.0', 'unknown'])
def test_versions(pattern):
    assert pattern not in infra_10x.__version__
    assert pattern not in py10x_infra.__version__



from infra_10x.mongodb_store import MongoStore


class MongodbAdmin:
    def __init__(self, hostname: str, username: str, password: str):
        self.client = MongoStore.connect(hostname=hostname, username=username, password=password, _cache=False)
        self.db = self.client[MongoStore.ADMIN]
        self.hostname = hostname
        self.username = username

    def user_exists(self, username: str) -> bool:
        res = self.db.command('usersInfo', {'user': username, 'db': MongoStore.ADMIN})
        return bool(res.get('users'))

    def update_user(self, username: str, password: str, *role_names, keep_current_roles=True):
        create = not self.user_exists(username)
        if create:
            if not password:
                raise RuntimeError(f'User {username} does not exist and requires a non-empty password')

            cmd = dict(createUser=username, pwd=password, roles=[*role_names])

        else:
            if not role_names and keep_current_roles:
                role_names = self.user_role_names(username)

            cmd = dict(updateUser=username)
            if role_names:
                cmd.update(roles=[*role_names])
            if password:
                cmd.update(pwd=password)

        self.db.command(cmd)

    def change_own_password(self, password: str):
        self.db.command(dict(updateUser=self.username, pwd=password))
        self.client = MongoStore.connect(hostname=self.hostname, username=self.username, password=password, _cache=False)
        self.db = self.client[MongoStore.ADMIN]

        # # Discover which user/db you're authenticated as (handy to avoid hardcoding)
        # status = client.admin.command("connectionStatus", showPrivileges = False)
        # auth = status[ "authInfo" ][ "authenticatedUsers" ]
        # if not auth:
        #     raise RuntimeError("Not authenticated. Connect with your current username/password first.")
        #
        # username = auth[0]['user']
        # user_db = auth[0]['db']
        #
        # # Change your password (roles remain unchanged because we don't pass 'roles')
        # client[user_db].command("updateUser", username, pwd = "<NEW_STRONG_PASSWORD>")
        #
        # # Reconnect with the new password (recommended)
        # client = MongoClient(f"mongodb://{username}:<NEW_STRONG_PASSWORD>@localhost:27017/?authSource={user_db}")
        # client.admin.command("ping")
        # print("Password changed and re-authenticated successfully.")

    def update_role(self, role_name: str, **permissions_per_db):
        """
        :param role_name: e.g., 'VISITOR'
        :param permissions_per_db: a dict: each key is either a db name or db_name/collection_name, value is PERMISSION, e.g.
                { 'catalog': PERMISSION.READ_ONLY, 'logs/errors': PERMISSION.READ_WRITE }
        :return:
        """
        privileges = [
            dict(resource=dict(db=resource_parts[0], collection=resource_parts[1] if len(resource_parts) == 2 else ''), actions=permission)
            for db_name, permission in permissions_per_db.items()
            if (resource_parts := db_name.split('/'))
        ]
        method = 'createRole' if not self.role_exists(role_name) else 'updateRole'
        self.db.command(method, role_name, privileges=privileges, roles=[])

    def all_user_names(self) -> list:
        listing = self.db.command('usersInfo')
        return [doc['user'] for doc in listing['users']] if listing['ok'] else []

    def all_role_names(self) -> list:
        listing = self.db.command('rolesInfo')
        return [doc['role'] for doc in listing['roles']] if listing['ok'] else []

    def role_exists(self, role_name: str) -> bool:
        res = self.db.command('rolesInfo', dict(role=role_name, db=MongoStore.ADMIN))
        return bool(res['roles'])

    def user_role_names(self, username: str) -> list:
        res = self.db.command(dict(usersInfo=username))
        if not res:
            return []

        return [r['role'] for r in res['users'][0]['roles']]

    def delete_all_roles(self):
        db = self.db
        for role_name in self.all_role_names():
            db.command('dropRole', role_name)

    def delete_role(self, role_name: str):
        if self.role_exists(role_name):
            self.db.command('dropRole', role_name)

    def delete_all_users(self):
        db = self.db
        me = self.username
        for username in self.all_user_names():
            if username != me:
                db.command('dropUser', username)

    def delete_user(self, username: str):
        me = self.username
        if me != username:
            if self.user_exists(username):
                self.db.command('dropUser', username)



from __future__ import annotations

from pathlib import Path
import importlib.metadata as md

import core_10x
from core_10x.global_cache import cache

PY10X_ROOT = Path(core_10x.__file__).resolve().parent.parent

@cache
def _owned_top_levels() -> set[str] | None:
    try:
        files = md.distribution('py10x-core').files or []
    except md.PackageNotFoundError:
        return None

    tops: set[str] = set()
    for f in files:
        p = Path(f)
        if not p.parts:
            continue
        top = p.parts[0]
        if Path(top).is_relative_to(PY10X_ROOT):
            tops.add(top)

    return tops


def pytest_ignore_collect(collection_path, config):
    """Only constrain collection for py10x package paths."""
    p = Path(collection_path).resolve()
    if not p.is_relative_to(PY10X_ROOT):
        # Returning None means "no opinion" so user package collection is unaffected.
        return None

    if any('.venv' in part for part in p.parts):
        return True

    tops = _owned_top_levels()
    rel = p.relative_to(PY10X_ROOT)
    parts = rel.parts
    if not parts:
        return False

    if tops and parts[0] not in tops:
        return True

    if rel.is_dir():
        return False

    # Do not ignore tests located in a unit_tests parent directory.
    return not(len(parts) > 1 and parts[-2] == 'unit_tests')




from core_10x.environment_variables import EnvVars
from core_10x.global_cache import cache
from core_10x.ts_store import TsStore


class BackboneStore:
    """
    Backbone Store instance should exist for every build area. It may be run with or without authentication required.
    If authentication is required, it must have a special-purpose user SIDEKICK with read-only permissions and must have the following databases:
    - admin
    - _backbone_    - collections:
        - bound data domains;
        - users;
        - user groups;
        - ...
    - _vault_       - collections of Resource Accessors (per username)
    - _shadow_      - collections of 'Shadow' (user-password-encrypted private key) for each user (for exporting sec keys for a new user's computer)

    """

    # fmt: off
    SIDEKICK_USER       = 'sidekick'
    BACKBONE_DB_NAME    = '_backbone_'
    VAULT_DB_NAME       = '_vault_'
    SHADOW_DB_NAME      = '_shadow_'
    # fmt: on

    @classmethod
    @cache
    def bb_store(cls) -> TsStore:  # -- BB Store
        cls_name = EnvVars.backbone_store_class_name
        cls.hostname = hostname = EnvVars.backbone_store_host_name
        if not cls_name or not hostname:
            return None

        cls.store_cls = store_cls = TsStore.store_class(cls_name)

        is_running, with_auth = store_cls.is_running_with_auth(hostname)
        if not is_running:
            raise OSError(f'{store_cls.s_driver_name}({hostname}) is not running')

        cls.with_auth = with_auth
        if with_auth:
            cls.username = cls.SIDEKICK_USER
            cls.password = cls.SIDEKICK_USER
        else:
            cls.username = ''
            cls.password = ''

        return store_cls.instance(hostname=hostname, dbname=cls.BACKBONE_DB_NAME, username=cls.username, password=cls.password)

    @classmethod
    @cache
    def store(cls, db_name: str) -> TsStore:
        bb_store = cls.bb_store()
        if db_name == cls.BACKBONE_DB_NAME:
            return bb_store

        return cls.store_cls.instance(hostname=cls.hostname, dbname=db_name, username=cls.username, password=cls.password)



# from PyQt5.QtCore import QModelIndex, QPointF, QRect, QSize, Qt, QVariant
# from PyQt5.QtGui import QFont, QFontMetrics, QPainter, QPaintEvent
# from PyQt5.QtWidgets import QStyle, QStyleOptionHeader

from ui_10x.platform import ux
from ui_10x.traitable_view import TraitableView


class HeaderModel(ux.StandardItemModel):
    def _create_subtree(self, traits: dict, root: ux.StandardItem, tree: dict):
        col = 0
        for subtree_name, subtree in tree.items():
            if isinstance(subtree, str):
                trait = traits.get(subtree_name)
                if not trait:
                    continue
                label = subtree
                subtree = None
            elif isinstance(subtree, dict):
                label = subtree_name
            else:
                continue

            item = ux.StandardItem(label)
            if subtree:
                self._create_subtree(traits, item, subtree)

            if root:
                root.append_column((item, ))
            else:
                self.set_item(0, col, item)

            col += 1

    def __init__(self, traits: dict, header_structure: dict):
        super().__init__()
        self.traits = traits
        self.header_structure = header_structure
        self.create(header_structure)

    def create(self, header_structure: dict):
        self._create_subtree(self.traits, None, header_structure)

    def leaf_index(self, section_index: int) -> ux.ModelIndex:
        current_leaf_index = -1
        for i in range(self.column_count()):
            res, current_leaf_index = self.find_leaf(self.index( 0, i), section_index, current_leaf_index)
            if res.is_valid():
                return res

        return ux.ModelIndex()

    @staticmethod
    def leftmost_leaf(index: ux.ModelIndex) -> ux.ModelIndex:
        prev_index = index
        while index.is_valid():
            prev_index = index
            index = index.child(0, 0)
        return prev_index

    @staticmethod
    def find_leaf(current_index: ux.ModelIndex, section_index: int, current_leaf_index: int) -> tuple:  #-- ( index, current_leaf_index )
        if current_index.is_valid():
            child_count = current_index.model().column_count(current_index)
            if child_count:
                for i in range(child_count):
                    res, current_leaf_index = HeaderModel.find_leaf(current_index.child(0, i), section_index, current_leaf_index)
                    if res.is_valid():
                        return (res, current_leaf_index)
            else:
                current_leaf_index += 1
                if current_leaf_index == section_index:
                    return (current_index, current_leaf_index)

        return (ux.ModelIndex(), current_leaf_index)

    @staticmethod
    def search_leaves(current_index: ux.ModelIndex) -> list:
        res = []
        if current_index.is_valid():
            if not current_index.child(0, 0).is_valid():
                res.append(current_index)
            else:
                c = 0
                child = current_index.child(0, c)
                while child.is_valid():
                    found = HeaderModel.search_leaves(child)
                    res.extend(found)
                    c += 1
                    child = current_index.child(0, c)

        return res

    @staticmethod
    def leaves(search_index: ux.ModelIndex ) -> list:
        leaves = []
        if search_index.is_valid():
            child_count = search_index.model().column_count(search_index)
            for i in range(child_count):
                found = HeaderModel.search_leaves(search_index.child(0, i))
                leaves.extend(found)

        return leaves

    @staticmethod
    def find_root_index(index: ux.ModelIndex) -> ux.ModelIndex:
        while index.parent().is_valid():
            index = index.parent()
        return index

    @staticmethod
    def parent_indexes(index: ux.ModelIndex) -> list:
        indexes = []
        while index.is_valid():
            indexes.insert(0, index)
            index = index.parent()
        return indexes


class HeaderView(ux.HeaderView):
    def __init__(self, view: TraitableView, parent: ux.Widget = None):
        super().__init__(ux.Horizontal, parent = parent)
        self.m_model = HeaderModel(view.ui_hints, view.header)
        self.set_model(self.m_model)

        #self.setStretchLastSection(False)
        self.setSectionResizeMode(ux.HeaderView.ResizeMode.Stretch)

        self.painted_cells = set()
        ##self.section_resized_connect(self.on_section_resized)

    def set_foreground_brush(self, opt: ux.StyleOptionHeader, index: ux.ModelIndex ):
        fgb = index.data(ux.ForegroundRole)
        if fgb:
            opt.palette.set_brush(ux.Palette.ButtonText, fgb)

    def set_background_brush(self, opt: ux.StyleOptionHeader, index: ux.ModelIndex ):
        bgb = index.data(ux.BackgroundRole)
        if bgb:
            opt.palette.set_brush(ux.Palette.Button, bgb)
            opt.palette.set_brush(ux.Palette.Window, bgb)


    # s_selected_position = (
    #     ux.StyleOptionHeader.NotAdjacent,                     # prev - 0, next 0
    #     ux.StyleOptionHeader.NextIsSelected,                  # prev - 0, next 1
    #     ux.StyleOptionHeader.PreviousIsSelected,              # prev - 1, next 0
    #     ux.StyleOptionHeader.NextAndPreviousAreSelected       # prev - 1, next 1
    # )
    def style_option_for_cell(self, logical_index: int) -> ux.StyleOptionHeader:
        opt = ux.StyleOptionHeader()
        # self.init_style_option(opt)
        # if self.window().is_active_window():
        #     opt.state |= ux.Style.State_Active
        # opt.text_alignment = ux.AlignCenter
        # opt.icon_alignment = ux.AlignVCenter
        # opt.section = logical_index
        #
        # visual = self.visual_index(logical_index)
        #
        # if self.count() == 1:
        #     opt.position = ux.StyleOptionHeader.OnlyOneSection
        # else:
        #     if visual == 0:
        #         opt.position = QStyleOptionHeader.Beginning
        #     else:
        #         opt.position = QStyleOptionHeader.End if visual == ( self.count() - 1 ) else QStyleOptionHeader.Middle
        #
        # if self.sectionsClickable():
        #     if self.highlightSections() and self.selectionMode():
        #         if self.selectionModel().columnIntersectsSelection( logical_index, self.rootIndex() ):
        #             opt.state |= QStyle.State_On
        #         if self.selectionModel().isColumnSelected( logical_index, self.rootIndex() ):
        #             opt.state |= QStyle.State_Sunken
        #
        # if self.selectionModel():
        #     prev_selected = self.selectionModel().isColumnSelected( self.logicalIndex( visual - 1 ), self.rootIndex() )
        #     next_selected = self.selectionModel().isColumnSelected( self.logicalIndex( visual + 1 ), self.rootIndex() )
        #     code = 2 * prev_selected + next_selected
        #     opt.selectedPosition = self.s_selectedPosition[ code ]

        return opt

    # def cell_size(self, leaf_index: ux.ModelIndex, style_options: ux.StyleOptionHeader) -> ux.Size:
    #     res = ux.Size()
    #     variant = QVariant(leaf_index.data(Qt.SizeHintRole))
    #     if variant.isValid():
    #         res = variant   # TODO: cast?
    #     fnt = QFont(self.font())
    #     var = QVariant(leaf_index.data(Qt.FontRole))
    #     if var.isValid():
    #         fnt = var
    #     fnt.setBold(True)
    #     fm = QFontMetrics(fnt)
    #     size = QSize(fm.size(0, leaf_index.data(Qt.DisplayRole)))
    #     if leaf_index.data(Qt.UserRole):      # isValid()
    #         size.transpose()
    #     decoration_size = self.style().sizeFromContents(QStyle.CT_HeaderSection, style_options, QSize(), self)
    #     empty_text_size = QSize(fm.size( 0, '' ))
    #     return res.expandedTo(size + decoration_size - empty_text_size)
    #
    # def current_cell_width(self, searched_index: QModelIndex, leaf_index: QModelIndex, section_index: int) -> int:
    #     leaves = HeaderModel.leaves(searched_index)
    #     if not leaves:
    #         return self.section_size(section_index)
    #
    #     first_leaf_section_index = section_index - leaves.index(leaf_index)
    #     return sum(self.section_size(first_leaf_section_index + i) for i in range(len(leaves)))
    #
    # def current_cell_left(self, searched_index: QModelIndex, leaf_index: QModelIndex, section_index: int, left: int) -> int:
    #     leaves = HeaderModel.leaves(searched_index)
    #     if leaves:
    #         n = leaves.index(leaf_index)
    #         first_leaf_section_index = section_index - n
    #         n -= 1
    #         left -= sum(self.section_size(first_leaf_section_index + i) for i in range(n))
    #     return left
    #
    # def paint_horizontal_cell(self, painter: QPainter, cell_index: QModelIndex, leaf_index: QModelIndex, logical_index: int, style_options: QStyleOptionHeader, section_rect: QRect, top: int) -> int:
    #     uniopt = QStyleOptionHeader(style_options)
    #     self.set_foreground_brush(uniopt, cell_index)
    #     self.set_background_brush(uniopt, cell_index)   # TODO!
    #
    #     height = self.cell_size(cell_index, uniopt).height()
    #
    #     if cell_index == leaf_index:
    #         height = section_rect.height() - top
    #
    #     left    = self.current_cell_left(cell_index, leaf_index, logical_index, section_rect.left())
    #     width   = self.current_cell_width(cell_index, leaf_index, logical_index)
    #
    #     if cell_index not in self.painted_cells:
    #         self.painted_cells.add(cell_index)
    #         r = QRect(left, top, width, height)
    #
    #         uniopt.text = cell_index.data(Qt.DisplayRole)
    #         painter.save()
    #         uniopt.rect = r
    #         self.style().drawControl(QStyle.CE_Header, uniopt, painter, self)
    #
    #         painter.restore()
    #
    #     return top + height
    #
    # def paintEvent(self, event: QPaintEvent):
    #     self.painted_cells = set()
    #     super().paintEvent(event)
    #
    # def paintSection(self, painter: QPainter, rect: QRect, logical_index: int):
    #     if rect.isValid():
    #         leaf_index = self.model.leaf_index(logical_index)
    #         if leaf_index.is_valid():
    #             old_bo = QPointF(painter.brushOrigin())
    #             top = rect.y()
    #             indexes = HeaderModel.parent_indexes(leaf_index)
    #             style_options = self.style_option_for_cell(logical_index)
    #             for i in range(len(indexes)):
    #                 real_style_options = QStyleOptionHeader(style_options)
    #                 state = int(real_style_options.state)
    #                 if i < (len(indexes) - 1) and (bool(state & QStyle.State_Sunken) or bool(state & QStyle.State_On)):
    #                     real_style_options.state &= ~(QStyle.State_Sunken | QStyle.State_On)
    #
    #                 cell = indexes[i]
    #                 top = self.paint_horizontal_cell(painter, cell, leaf_index, logical_index, real_style_options, rect, top)
    #
    #             painter.setBrushOrigin(old_bo)
    #             return
    #
    #     super().paintSection(painter, rect, logical_index)
    #
    # def sectionSizeFromContents(self, logical_index: int ) -> QSize:
    #     cur_leaf_index = QModelIndex(self.model.leaf_index(logical_index))
    #     if not cur_leaf_index.is_valid():
    #         return super().sectionSizeFromContents(logical_index)
    #
    #     style_option = QStyleOptionHeader(self.style_option_for_cell(logical_index ))
    #     s = QSize(self.cell_size(cur_leaf_index, style_option))
    #     cur_leaf_index = cur_leaf_index.parent()
    #     while cur_leaf_index.is_valid():
    #         s.setHeight(s.height() + self.cell_size(cur_leaf_index, style_option).height())
    #         cur_leaf_index = cur_leaf_index.parent()
    #
    #     return s
    #
    # def on_section_resized(self, logical_index: int, i2, i3 ):
    #     if self.is_section_hidden(logical_index):
    #         return
    #
    #     leaf_index = QModelIndex(self.model.leaf_index(logical_index))
    #     if leaf_index.is_valid():
    #         root_index = HeaderModel.find_root_index(leaf_index)
    #         leaves = HeaderModel.leaves(root_index)
    #         try:
    #             n = leaves.index(leaf_index)
    #             while n > 0:
    #                 logical_index -= 1
    #
    #                 w = self.viewport().width()
    #                 h = self.viewport().height()
    #                 pos = self.sectionViewportPosition(logical_index)
    #                 r = QRect(pos, 0, w - pos, h)
    #                 if self.isRightToLeft():
    #                     r.setRect(0, 0, pos + self.sectionSize(logical_index), h)
    #
    #                 self.viewport().update( r.normalized() )
    #                 n -= 1
    #         except Exception:
    #             pass



from __future__ import annotations

import ui_10x.platform_interface as i
import ui_10x.rio.components as rio_components
from ui_10x.rio.component_builder import Widget


class CheckBox(Widget, i.CheckBox):
    s_component_class = rio_components.LabeledCheckBox

    def set_checked(self, checked: bool):
        self['is_on'] = checked

    def is_checked(self) -> bool:
        return self['is_on']

    def state_changed_connect(self, bound_method):
        def state_change_handler(event):
            bound_method(event.is_on)

        self['on_change'] = self.callback(state_change_handler)

    def set_text(self, text: str):
        self['label'] = text



# from PyQt6.QtWidgets import QWidget, QLayout, QBoxLayout, QVBoxLayout, QHBoxLayout, QFormLayout, QApplication
# from PyQt6.QtWidgets import QLabel, QCalendarWidget, QMessageBox, QGroupBox, QButtonGroup, QRadioButton
# from PyQt6.QtWidgets import QListWidgetItem, QListWidget, QLineEdit, QPlainTextEdit, QTreeWidgetItem, QTreeWidget, QCheckBox, QComboBox
# from PyQt6.QtWidgets import QFrame, QSplitter, QDialog, QScrollArea, QPushButton, QStyle, QSizePolicy
# from PyQt6.QtCore import Qt, QObject, QThread, pyqtSignal, QBuffer, QIODevice, QPoint    #, QByteArray
# from PyQt6.QtGui import QColor, QGuiApplication, QPixmap, QMouseEvent
#
# import platform
#
#
# Object = QObject
#
# signal_decl = pyqtSignal
#
# class MouseEvent(QMouseEvent):
#     def is_left_button(self) -> bool:   return self.button() == Qt.MouseButton.LeftButton
#     def is_right_button(self) -> bool:  return self.button() == Qt.MouseButton.RightButton

from ui_10x.platform import ux

if __name__ == '__main__':
    app = ux.init()

    w = ux.Label('a')
    w = ux.PushButton('b')
    w = ux.LineEdit()
    w = ux.RadioButton()
    w = ux.CheckBox('Check')
    w = ux.ListWidget()
    w = ux.TreeWidget()





